"""
"""


#[

from __future__ import annotations

import functools as _ft
import numpy as _np
import documark as _dm
from typing import Literal

from ._functionalize import FUNC_STRING

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import Any, Callable, EllipsisType
    from collections.abc import Iterable
    from numbers import Real
    from ..dates import Period
    from ..series import Series

#]


__all__ = []


MethodType = Literal["next", "previous", "nearest", "linear", "log_linear", "constant", "from_series", ]


class Inlay:
    """
    """
    #[

    @_dm.reference(category="homogenizing", )
    def fill_missing(
        self,
        method: Literal["next", "previous", "nearest", "linear", "log_linear", "constant", "from_series", ],
        method_args: Any | None = None,
        span: Iterable[Period] | EllipsisType | None = None,
    ) -> None:
        r"""
················································································

==Fill missing observations==


### Function form for creating new time `Series` objects ###

    new = irispie.fill_missing(
        self,
        method,
        *args,
        span=None,
    )


### Class method form changing existing `Series` objects in-place ###

    self.fill_missing(
        method,
        *args,
        span=None,
    )


### Input arguments ###


???+ input "self"
    The time `Series` object to be filled.

???+ input "method"
    The method to be used for filling missing observations. The following methods are available:

    | Method         | Description
    |----------------|-------------
    | "constant"     | Fill with a constant value
    | "next"         | Next available observation
    | "previous"     | Previous available observation
    | "nearest"      | Nearest available observation
    | "linear"       | Linear interpolation or extrapolation
    | "log_linear"   | Log-linear interpolation or extrapolation
    | "from_series"  | Fill with values from another time series object

???+ input "*args"
    Additional arguments to be passed to the filling method. The following methods require additional arguments:

    | Method         | Additional argument(s)
    |----------------|-----------------------
    | "constant"     | A single constant value
    | "from_series"  | A time `Series` object


???+ input "span"
    The time span to be filled. If `None`, the time span of the input time `Series` is filled.


### Returns ###


???+ returns "self"
    The time `Series` object with missing observations filled.

???+ returns "new"
    A new time `Series` object with missing observations filled.

················································································
        """
        fill_func = _FILL_METHOD_DISPATCH[method]
        data, span = self.get_data_and_periods(span, )
        new_data = [
            fill_func(variant, method_args, span=span, ).T
            for variant in data.T
        ]
        self.set_data(span, new_data, )

    #]


attributes = (n for n in dir(Inlay) if not n.startswith("_"))
for n in attributes:
    code = FUNC_STRING.format(n=n, )
    exec(code, globals(), locals(), )
    __all__.append(n)


def _fill_neighbor(
    data,
    method_args: Any | None,
    func: Callable,
    **kwargs,
) -> _np.ndarray:
    """
    """
    index_nan = _np.isnan(data)
    if _np.all(index_nan):
        return data
    where_obs = _np.where(~index_nan)[0].astype(_np.float64, )
    if where_obs.size == 0:
        return data
    where_nan = _np.where(index_nan)[0]
    for i in where_nan:
        j = func(i, where_obs, )
        data[i] = data[j] if j is not None else _np.nan
    return data


def _fill_interp(
    data: _np.ndarray,
    method_args: Any | None,
    func: Callable,
    **kwargs,
) -> _np.ndarray:
    """
    """
    index_nan = _np.isnan(data)
    if _np.all(index_nan):
        return data
    where_obs = _np.where(~index_nan)[0].astype(_np.float64, )
    if where_obs.size == 0:
        return data
    where_nan = _np.where(index_nan)[0]
    for i in where_nan:
        prev = _previous_index(i, where_obs, )
        next_ = _next_index(i, where_obs, )
        if prev is not None and next_ is not None:
            data[i] = func(data[prev], data[next_], prev, next_, i, )
            continue
        if prev is not None:
            data[i] = data[prev]
            continue
        if next_ is not None:
            data[i] = data[next_]
            continue
    return data


def _interpolation_linear(
    previous_value: Real,
    next_value: Real,
    previous_index: int,
    next_index: int,
    curr_index: int,
    /,
) -> Real:
    """
    """
    diff = next_value - previous_value
    scale_diff = (curr_index - previous_index) / (next_index - previous_index)
    return previous_value + diff * scale_diff


def _interpolation_log_linear(
    previous_value: Real,
    next_value: Real,
    previous_index: int,
    next_index: int,
    curr_index: int,
    /,
) -> Real:
    """
    """
    return _np.exp(_interpolation_linear(
        _np.log(previous_value),
        _np.log(next_value),
        previous_index,
        next_index,
        curr_index,
    ))


def _fill_constant(
    data: _np.ndarray,
    method_args: Real,
    **kwargs,
) -> _np.ndarray:
    """
    """
    constant = method_args
    index_nan = _np.isnan(data)
    data[index_nan] = constant
    return data


def fill_from_series(
    values: _np.ndarray,
    method_args: Series,
    span: Iterable[Period],
) -> _np.ndarray:
    """
    """
    series = method_args
    fill_values = series.get_data(span, )
    index_nan = _np.isnan(values)
    values[index_nan] = fill_values.flatten()[index_nan]
    return values


def _next_index(i: int, where_obs: _np.ndarray, /, ) -> int:
    """
    """
    #[
    where_obs_minus_i = where_obs - i
    index_negative = where_obs_minus_i < 0
    if _np.all(index_negative):
        return None
    where_obs_minus_i[index_negative] = _np.inf
    index = _np.argmin(where_obs_minus_i)
    return int(where_obs[index])
    #]


def _previous_index(i: int, where_obs: _np.ndarray, /, ) -> int:
    """
    """
    #[
    i_minus_where_obs = i - where_obs
    index_negative = i_minus_where_obs < 0
    if _np.all(index_negative):
        return None
    i_minus_where_obs[index_negative] = _np.inf
    index = _np.argmin(i_minus_where_obs)
    return int(where_obs[index])
    #]


def _nearest_index(i: int, where_obs: _np.ndarray, /, ) -> int:
    """
    """
    #[
    i_minus_where_obs = _np.abs(i - where_obs)
    index = _np.argmin(i_minus_where_obs)
    return int(where_obs[index])
    #]


_FILL_METHOD_DISPATCH = {
    "next": _ft.partial(_fill_neighbor, func=_next_index, ),
    "previous": _ft.partial(_fill_neighbor, func=_previous_index, ),
    "nearest": _ft.partial(_fill_neighbor, func=_nearest_index, ),
    "linear": _ft.partial(_fill_interp, func=_interpolation_linear, ),
    "log_linear": _ft.partial(_fill_interp, func=_interpolation_log_linear, ),
    "constant": _fill_constant,
    "from_series": fill_from_series,
    #
    # Aliases for backward compatibility
    "series": fill_from_series,
}






import irispie as ir


def test_yearly():
    t = ir.yy(2023)
    assert t.to_sdmx_string() == "2023"


def test_halfyearly():
    t = ir.hh(2023, 1)
    assert t.to_sdmx_string() == "2023-H1"


def test_quarterly():
    t = ir.qq(2023, 1)
    assert t.to_sdmx_string() == "2023-Q1"


def test_monthly():
    t = ir.mm(2023, 1)
    assert t.to_sdmx_string() == "2023-01"


def test_daily():
    t = ir.dd(2023, 1, 25)
    assert t.to_sdmx_string() == "2023-01-25"




r"""
Reduced-form vector autoregression models
"""


#[

from __future__ import annotations

from typing import NoReturn

from .. import has_invariant as _has_invariant
from .. import has_variants as _has_variants
from .. import quantities as _quantities
from ..quantities import Quantity, QuantityKind
from ..fords import covariances as _covariances
from ..namings import DimensionNames

from . import _estimators as _estimators
from . import _simulators as _simulators
from . import _slatable_protocols as _slatable_protocols
from ._invariants import Invariant
from ._variants import Variant, System
from ._dimensions import Dimensions

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from collections.abc import Iterable
    from ..fords.descriptors import Solution, SolutionVectors

#]


__all__ = (
    "RedVAR",
)


class RedVAR(
    _estimators.Inlay,
    _simulators.Inlay,
    _slatable_protocols.Inlay,
    _has_invariant.Mixin,
    _has_variants.Mixin,
    _quantities.Mixin,
):
    r"""
    """
    #[

    __slots__ = (
        "_invariant",
        "_variants",
    )

    max_lead = 0

    def __init__(
        self,
        *args,
        num_variants: int = 1,
        **kwargs,
    ) -> None | NoReturn:
        """
        """
        if not args and not kwargs:
            self._invariant = None
            self._variants = []
            return
        self._invariant = Invariant(*args, **kwargs, )
        self._variants = [ Variant() for _ in range(num_variants, ) ]

    @classmethod
    def skeleton(
        klass,
        other: Self,
    ) -> Self:
        r"""
        """
        self = klass()
        self._invariant = other._invariant
        self._variants = []
        return self

    def copy(self, ) -> Self:
        r"""
        """
        new = type(self)()
        new._invariant = self._invariant
        new._variants = [ v.copy() for v in self._variants ]
        return new

    def get_endogenous_names(self, ) -> tuple[str, ...]:
        r"""
        """
        quantities = self._access_quantities()
        return tuple(_quantities.generate_names_of_kind(quantities, QuantityKind.TRANSITION_VARIABLE, ))

    def get_exogenous_names(self, ) -> tuple[str, ...]:
        r"""
        """
        quantities = self._access_quantities()
        return tuple(_quantities.generate_names_of_kind(quantities, QuantityKind.EXOGENOUS_VARIABLE, ))

    def get_residual_names(self, ) -> tuple[str, ...]:
        r"""
        """
        quantities = self._access_quantities()
        return tuple(_quantities.generate_names_of_kind(quantities, QuantityKind.UNANTICIPATED_SHOCK, ))

    def get_endogenous_qids(self, ) -> tuple[int, ...]:
        r"""
        """
        return self._invariant.get_endogenous_qids()

    def get_residual_qids(self, ) -> tuple[int, ...]:
        r"""
        """
        return self._invariant.get_residual_qids()

    def get_exogenous_qids(self, ) -> tuple[int, ...]:
        r"""
        """
        return self._invariant.get_exogenous_qids()

    def get_system_matrices(
        self,
        unpack_singleton: bool = True,
    ) -> System | tuple[System, ...]:
        r"""
        """
        system_matrices = [ v.system for v in self._variants ]
        return self.unpack_singleton(system_matrices, unpack_singleton=unpack_singleton, )

    def get_acov_dimension_names(self, ) -> DimensionNames:
        r"""
        """
        endogenous_names = self.get_endogenous_names()
        dimension_names = DimensionNames(rows=endogenous_names, columns=endogenous_names, )
        return dimension_names

    def get_acov(
        self,
        up_to_order: int = 0,
        unpack_singleton: bool = True,
    ) -> tuple[_np.ndarray, ...] | list[tuple[_np.ndarray, ...]]:
        r"""
        """
        acov_by_variant = [
            v.get_acov(up_to_order=up_to_order, )
            for v in self._variants
        ]
        return self.unpack_singleton(
            acov_by_variant,
            unpack_singleton=unpack_singleton,
        )

    def get_acorr(self, *args, **kwargs, ):
        r"""
        """
        return _covariances.get_acorr_by_variant(self, *args, **kwargs, )

    def get_mean(
        self,
        unpack_singleton: bool = True,
    ) -> _np.ndarray | list[_np.ndarray]:
        r"""
        """
        amean_by_variant = [ v.get_mean() for v in self._variants ]
        return self.unpack_singleton(
            amean_by_variant,
            unpack_singleton=unpack_singleton,
        )

    @property
    def dimensions(self, ) -> Dimensions:
        r"""
        """
        return self._invariant.dimensions

    @property
    def order(self, ) -> int:
        r"""
        """
        return self._invariant.dimensions.order

    @property
    def max_lag(self, ) -> int:
        r"""
        """
        return self._invariant.dimensions.order

    @property
    def has_intercept(self, ) -> bool:
        r"""
        """
        return self._invariant.dimensions.has_intercept

    @property
    def has_exogenous(self, ) -> int:
        r"""
        """
        return self._invariant.dimensions.has_exogenous

    def _get_dynamic_solution_vectors(self, ) -> SolutionVectors:
        r"""
        """
        return self._invariant.solution_vectors

    def _get_singleton_solution(
        self,
        vid: int = 0,
        **kwargs,
    ) -> Solution:
        r"""
        """
        return self._variants[0]._get_companion_solution(**kwargs, )

    def get_companion_matrices(
        self,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> tuple[System, ...] | System:
        r"""
        """
        companion_matrices = [
            v._get_companion_solution(**kwargs, )
            for v in self._variants
        ]
        return self.unpack_singleton(
            companion_matrices,
            unpack_singleton=unpack_singleton,
        )

    def get_stability(
        self,
        unpack_singleton: bool = True,
    ) -> list[bool, ...] | bool:
        r"""
        """
        stability = [ v.is_stable for v in self._variants ]
        return self.unpack_singleton(
            stability,
            unpack_singleton=unpack_singleton,
        )

    def get_max_abs_eigenvalue(
        self,
        unpack_singleton: bool = True,
    ) -> list[bool, ...] | bool:
        r"""
        """
        max_abs_eigenvalue = [ v.max_abs_eigenvalue for v in self._variants ]
        return self.unpack_singleton(
            max_abs_eigenvalue,
            unpack_singleton=unpack_singleton,
        )

    def get_eigenvalues(
        self,
        unpack_singleton: bool = True,
    ) -> list[bool, ...] | bool:
        r"""
        """
        eigenvalues = [ v.eigenvalues for v in self._variants ]
        return self.unpack_singleton(
            eigenvalues,
            unpack_singleton=unpack_singleton,
        )

#]




"""
Plain equator
"""


#[
from __future__ import annotations

import numpy as _np

from .. import equations as _equations
from .. import makers as _makers

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any, Callable
    from collections.abc import Iterable
#]


EQUATOR_ARGS = ("x", "t", )


class PlainEquator:
    """
    """
    #[

    _state_slots = (
        "_context",
        "_equations",
        "_func_str",
        "_columns",
        "min_shift",
        "max_shift",
    )

    _nonstate_slots = (
        "_func",
    )

    __slots__ = _state_slots + _nonstate_slots

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        columns: Iterable[int] | None = None,
        context: dict[str, Callable] | None = None,
    ) -> None:
        """
        """
        self._equations = tuple(equations)
        self._populate_min_max_shifts()
        self._populate_columns(columns, )
        self._context = context
        self._create_function()

    @property
    def humans(self, /, ) -> tuple[str]:
        return tuple(e.human for e in self._equations)

    @property
    def num_equations(self, /, ) -> int:
        """
        """
        return len(self._equations)

    def _create_function(self, /, ) -> None:
        """
        """
        joined_xtrings = "  ,  ".join(i.xtring for i in self._equations)
        expression = "(" + joined_xtrings + " , )"
        self._func, self._func_str, *_ \
            = _makers.make_function("__equator", EQUATOR_ARGS, expression, self._context, )

    def _populate_min_max_shifts(self, /, ) -> None:
        """
        """
        self.min_shift = _equations.get_min_shift_from_equations(self._equations, )
        self.max_shift = _equations.get_max_shift_from_equations(self._equations, )

    def _populate_columns(self, columns: Iterable[int] | None, /, ) -> None:
        """
        """
        if columns is None:
            self._columns = None
        elif isinstance(columns, int):
            self._columns = columns
        else:
            self._columns = _np.array(columns, dtype=int, )

    @property
    def min_num_columns(self, /, ) -> int:
        return -self.min_shift + 1 + self.max_shift

    def eval(
        self,
        data_array: _np.ndarray,
        columns: int | _np.ndarray | None = None,
    ) -> _np.ndarray:
        """
        """
        if columns is None:
            columns = self._columns
        return self._func(data_array, columns, )

    def eval_as_array(
        self,
        *args,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        value = self.eval(*args, **kwargs, )
        return _np.array(value, dtype=_np.float64, ).reshape(self.num_equations, -1, )

    def __getstate__(self, ) -> dict[str, Any]:
        """
        """
        return {
            k: getattr(self, k)
            for k in self._state_slots
        }

    def __setstate__(self, state: dict[str, Any], ) -> None:
        """
        """
        for n in self._state_slots:
            setattr(self, n, state[n], )
        self._create_function()

    #]




"""
"""


#[

from __future__ import annotations

from typing import Protocol
from typing import Literal
import functools as _ft
import numpy as _np
import documark as _dm

from .. import has_variants as _has_variants
from .. import quantities as _quantities
from ..frames import Frame
from ..dataslates.main import Dataslate
from ..databoxes.main import Databox
from ..series.main import Series
from .solutions import Solution, right_div, left_div

from . import initializers as _initializers
from . import shock_simulators as _shock_simulators
from . import covariances as _covariances
from .descriptors import Squid
from .. import wrongdoings as _wd

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any, Iterable, Sequence, Self, NoReturn
    from types import EllipsisType
    from numbers import Real
    from ..dates import Period

#]


_simulate_anticipated_shocks \
    = _shock_simulators.simulate_triangular_anticipated_shocks


_DEFAULT_DELTA_TOLERANCE = 1e-12


DiffuseMethodType = Literal[
    "approx_diffuse",
    "fixed_unknown",
    "fixed_zero",
]


class SingularMatrixError(ValueError, ):
    pass


class KalmanFilterableProtocol(Protocol, ):
    """
    """
    #[

    is_singleton: bool

    def _get_singleton_solution(self, deviation: bool = False, ) -> Solution: ...

    def get_cov_unanticipated_shocks(self, ) -> _np.ndarray: ...

    #]


class KalmanOutputData:
    """
......................................................................

==Output data from Kalman filter==


......................................................................
    """
    #[

    __slots__ = (
        "predict_mse_obs",
        "predict_med",
        "predict_std",
        "predict_err",
        "update_med",
        "update_std",
        "smooth_med",
        "smooth_std",
    )

    def __init__(
        self: Self,
        output_store: _OutputStore,
    ) -> None:
        """
        """
        for attr_name in self.__slots__:
            attr = getattr(output_store, attr_name, )
            if hasattr(attr, "to_output_arg", ):
                self_attr = attr.to_output_arg()
            else:
                self_attr = attr
            setattr(self, attr_name, self_attr, )

    def __repr__(self, /, ) -> str:
        """
        """
        slots_to_repr = (
            i for i in self.__slots__
            if getattr(self, s) is not None
        )
        return f"KalmanOutputData({', '.join(slots_to_repr)})"

    #]


class Needs:
    """
    """
    #[

    __slots__ = (
        "return_predict",
        "return_update",
        "return_smooth",
        "return_predict_err",
        "return_predict_mse_obs",
        "rescale_variance",
        "likelihood_contributions",
        "output_store",
    )

    def __init__(
        self: Self,
        return_: Iterable[str, ...] = (),
        return_predict: bool = False,
        return_update: bool = False,
        return_smooth: bool = False,
        return_predict_err: bool = False,
        return_predict_mse_obs: bool = False,
        rescale_variance: bool = False,
        likelihood_contributions: bool = False,
    ) -> None:
        """
        """
        return_ = tuple(return_) if not isinstance(return_, str) else (return_, )
        self.return_predict = return_predict and "predict" in return_
        self.return_update = return_update and "update" in return_
        self.return_smooth = return_smooth and "smooth" in return_
        self.return_predict_err = return_predict_err and "predict_err" in return_
        self.return_predict_mse_obs = return_predict_mse_obs and "predict_mse_obs" in return_
        self.rescale_variance = rescale_variance
        self.likelihood_contributions = likelihood_contributions
        #
        self.output_store = any((
            self.return_predict,
            self.return_update,
            self.return_smooth,
        ))

    #]


class _LogDataslate:
    """
    """
    #[

    @classmethod
    def from_dataslate(
        klass,
        input_ds: Dataslate,
        num_variants: int,
        name_to_log_name: dict[str, str],
    ) -> Self | None:
        """
        """
        self = klass()
        self._dataslate = Dataslate.nan_from_template(input_ds, num_variants=num_variants, )
        self._prepare_log_names(name_to_log_name, )
        return self

    @classmethod
    def from_names_dates(
        klass,
        measurement_names: Iterable[str],
        dates: Iterable[Period],
        num_variants: int,
        name_to_log_name: dict[str, str],
    ) -> Self:
        """
        """
        self = klass()
        self._dataslate = Dataslate.nan_from_names_periods(
            measurement_names, dates,
            num_variants=num_variants,
        )
        self._prepare_log_names(name_to_log_name, )
        return self

    def store(
        self: Self,
        array: _np.ndarray,
        lhs_indexes: tuple,
        rhs_indexes=None,
        transform=None,
    ) -> None:
        """
        """
        if array is None:
            return
        rhs_indexes = rhs_indexes if rhs_indexes is not None else slice(None, )
        if transform is not None:
            array = transform[rhs_indexes, :] @ array
        else:
            array = array[rhs_indexes]
        self._dataslate._variants[0].data[lhs_indexes] = array

    def _prepare_log_names(
        self,
        name_to_log_name: dict[str, str],
        /,
    ) -> None:
        """
        """
        self._dataslate.rename(name_to_log_name, )
        output_names = self._dataslate.output_names
        self._name_to_log_name = {
            name: log_name
            for name, log_name in name_to_log_name.items()
            if log_name in output_names
        }

    def extend(
        self,
        other: Self,
        /,
    ) -> None:
        """
        """
        self._dataslate.extend(other._dataslate, )

    #]


class _MedLogDataslate(_LogDataslate, ):
    """
    """
    #[

    def to_output_arg(
        self: Self,
        /,
    ) -> Databox:
        """
        """
        db = self._dataslate.to_databox()
        for name, log_name in self._name_to_log_name.items():
            db[name] = db[log_name].copy()
            db[name].exp()
        return db

    #]


class _StdLogDataslate(_LogDataslate, ):
    """
    """
    #[

    def store_from_mse(
        self: Self,
        mse_array: _np.ndarray,
        lhs_indexes: tuple,
        rhs_indexes=None,
        transform=None,
    ) -> None:
        if mse_array is None:
            return
        rhs_indexes = rhs_indexes if rhs_indexes is not None else slice(None, )
        if transform is not None:
            transform_rhs_indexes = transform[rhs_indexes, :]
            mse_array = transform_rhs_indexes @ mse_array @ transform_rhs_indexes.T
        else:
            mse_array = mse_array[rhs_indexes, ...][..., rhs_indexes]
        self._dataslate._variants[0].data[lhs_indexes] = _std_from_mse(mse_array, )

    def to_output_arg(
        self: Self,
        /,
    ) -> Databox:
        """
        """
        return self._dataslate.to_databox()

    #]


class _OutputStore:
    """
    """
    #[

    __slots_to_input__ = (
        "transform",
        "squid",
        "measurement_incidence",
    )

    __slots_to_create__ = (
        "predict_med",
        "predict_std",
        "predict_mse_obs",

        "update_med",
        "update_std",
        "predict_err",

        "smooth_med",
        "smooth_std",
    )

    __slots__ = __slots_to_input__ + __slots_to_create__

    _OUT_NAMES = (
        "predict_mse_obs",
        "predict_med",
        "predict_std",
        "predict_err",
        "update_med",
        "update_std",
        "smooth_med",
        "smooth_std",
    )

    def __init__(
        self,
        input_ds: Dataslate,
        num_variants: int,
        measurement_names: Iterable[str],
        needs: Needs,
        name_to_log_name: dict[str, str] | None = None,
        **kwargs,
    ) -> Self:
        """
        """
        #
        for n in self.__slots_to_input__:
            setattr(self, n, kwargs.get(n, ), )
        #
        _med_constructor = _ft.partial(
            _MedLogDataslate.from_dataslate,
            input_ds=input_ds,
            num_variants=num_variants,
            name_to_log_name=name_to_log_name,
        )
        #
        _std_constructor = _ft.partial(
            _StdLogDataslate.from_dataslate,
            input_ds=input_ds,
            num_variants=num_variants,
            name_to_log_name=name_to_log_name,
        )
        #
        # Prediction step
        #
        self.predict_med = _med_constructor() if needs.return_predict else None
        self.predict_std = _std_constructor() if needs.return_predict else None
        self.predict_mse_obs = [
            [None]*input_ds.num_periods for _ in range(num_variants, )
        ] if needs.return_predict and needs.return_predict_mse_obs else None
        #
        # Updating step
        #
        self.update_med = _med_constructor() if needs.return_update else None
        self.update_std = _std_constructor() if needs.return_update else None
        self.predict_err = _MedLogDataslate.from_names_dates(
            measurement_names,
            input_ds.periods,
            num_variants,
            name_to_log_name,
        ) if needs.return_predict_err else None
        #
        # Smoothing step
        #
        self.smooth_med = _med_constructor() if needs.return_smooth else None
        self.smooth_std = _std_constructor() if needs.return_smooth else None

    def extend(self, other: Self, /, ) -> None:
        """
        """
        for s in self.__slots__:
            if getattr(self, s) is not None:
                getattr(self, s).extend(getattr(other, s), )

    def rescale_stds(self, var_scale, /, ) -> None:
        r"""
        """
        if var_scale == 1 or var_scale is None:
            return
        std_scale = _sqrt_positive(var_scale)
        for attr_name in ("predict_std", "update_std", "smooth_std", ):
            attr = getattr(self, attr_name, )
            if attr is None:
                continue
            for v in attr._dataslate._variants:
                v.rescale_data(std_scale)

    def create_out_data(self: Self, /, ) -> Databox:
        r"""
        """
        out_data = Databox()
        for name in self._OUT_NAMES:
            attr = getattr(self, name, )
            if attr is None:
                continue
            if hasattr(attr, "to_output_arg", ):
                out_data[name] = attr.to_output_arg()
            else:
                out_data[name] = attr
        return out_data

    def store_predict(
        self: Self,
        t: int | EllipsisType,
        inx_y: _np.ndarray | None = None,
        a0: _np.ndarray | None = None,
        u0: _np.ndarray | None = None,
        v0: _np.ndarray | None = None,
        y0: _np.ndarray | None = None,
        w0: _np.ndarray | None = None,
        Q0: _np.ndarray | None = None,
        F: _np.ndarray | None = None,
        cov_u0: _np.ndarray | None = None,
        cov_w0: _np.ndarray | None = None,
    ) -> None:
        r"""
        """
        if self.predict_med is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            u_qids = self.squid.u_qids
            v_qids = self.squid.v_qids
            w_qids = self.squid.w_qids
            y_qids = self.squid.y_qids
            #
            self.predict_med.store(a0, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
            self.predict_med.store(u0, (u_qids, t), )
            self.predict_med.store(v0, (v_qids, t), )
            self.predict_med.store(w0, (w_qids, t), )
            full_y0 = self._expand_y_to_full(y0, t, )
            self.predict_med.store(full_y0, (y_qids, t), )
            #
        if self.predict_std is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            u_qids = self.squid.u_qids
            w_qids = self.squid.w_qids
            self.predict_std.store_from_mse(Q0, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
            self.predict_std.store_from_mse(cov_u0, (u_qids, t), )
            self.predict_std.store_from_mse(cov_w0, (w_qids, t), )
            #
        if self.predict_mse_obs is not None and F is not None:
            self.predict_mse_obs[0][t] = F

    def store_update(
        self: Self,
        t: int | EllipsisType,
        inx_y: _np.ndarray | None = None,
        xi: _np.ndarray | None = None,
        u: _np.ndarray | None = None,
        v: _np.ndarray | None = None,
        w: _np.ndarray | None = None,
        y: _np.ndarray | None = None,
        pe: _np.ndarray | None = None,
        Q: _np.ndarray | None = None,
    ) -> None:
        """
        """
        if self.update_med is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            u_qids = self.squid.u_qids
            v_qids = self.squid.v_qids
            w_qids = self.squid.w_qids
            y_qids = self.squid.y_qids
            self.update_med.store(xi, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
            self.update_med.store(u, (u_qids, t), )
            self.update_med.store(v, (v_qids, t), )
            self.update_med.store(w, (w_qids, t), )
            full_y = self._expand_y_to_full(y, t, )
            self.update_med.store(full_y, (y_qids, t), )
        if self.update_std is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            self.update_std.store_from_mse(Q, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
        if self.predict_err is not None:
            full_pe = self._expand_y_to_full(pe, t, )
            self.predict_err.store(full_pe, (..., t), )

    def store_smooth(
        self: Self,
        t: int | EllipsisType,
        inx_y: _np.ndarray | None = None,
        xi: _np.ndarray | None = None,
        u: _np.ndarray | None = None,
        v: _np.ndarray | None = None,
        w: _np.ndarray | None = None,
        y: _np.ndarray | None = None,
        Q: _np.ndarray | None = None,
    ) -> None:
        """
        """
        if self.smooth_med is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            u_qids = self.squid.u_qids
            v_qids = self.squid.v_qids
            w_qids = self.squid.w_qids
            y_qids = self.squid.y_qids
            self.smooth_med.store(xi, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )
            self.smooth_med.store(u, (u_qids, t), )
            self.smooth_med.store(v, (v_qids, t), )
            self.smooth_med.store(w, (w_qids, t), )
            full_y = self._expand_y_to_full(y, t, )
            self.smooth_med.store(full_y, (y_qids, t), )
        if self.smooth_std is not None:
            curr_xi_qids = self.squid.curr_xi_qids
            curr_xi_indexes = self.squid.curr_xi_indexes
            self.smooth_std.store_from_mse(Q, (curr_xi_qids, t), rhs_indexes=curr_xi_indexes, transform=self.transform, )

    def _expand_y_to_full(
        self,
        observed: _np.ndarray | None,
        t: int | EllipsisType,
    ) -> _np.ndarray:
        """
        """
        if observed is None or t is Ellipsis:
            return observed
        inx_y = self.measurement_incidence[:, t]
        full = _np.full(inx_y.shape, _np.nan, )
        full[inx_y] = observed
        return full

    #]


def _sqrt_positive(x):
    return _np.sqrt(_np.maximum(x, 0.0, ))


def _std_from_mse(mse, ):
    return _sqrt_positive(_np.diag(mse).copy())


class Mixin:
    """
    """
    #[


    @_dm.reference(category="filtering", )
    def neg_log_likelihood(self, *args, **kwargs, ) -> Real:
        """
        """
        kwargs["return_"] = ()
        kwargs["likelihood_contributions"] = False
        kwargs["return_info"] = True
        _, info = self.kalman_filter(*args, **kwargs, )
        return info["neg_log_likelihood"]

    @_dm.reference(category="filtering", )
    def kalman_filter(
        self,
        #
        input_db: Databox,
        span: Iterable[Period],
        #
        diffuse_scale: Real | None = None,
        diffuse_method: DiffuseMethodType = "fixed_unknown",
        #
        return_: Iterable[str, ...] = ("predict", "update", "smooth", "predict_err", "predict_mse_obs", ),
        return_predict: bool = True,
        return_update: bool = True,
        return_smooth: bool = True,
        return_predict_err: bool = True,
        return_predict_mse_obs: bool = True,
        rescale_variance: bool = False,
        likelihood_contributions: bool = True,
        #
        shocks_from_data: bool = False,
        stds_from_data: bool = False,
        initials_from_data: bool = False,
        #
        prepend_initial: bool = False,
        append_terminal: bool = False,
        deviation: bool = False,
        check_singularity: bool = False,
        when_singularity: Literal["critical", "error", "warning", "silent" ] | None = "critical",
        num_variants: int | None = None,
        #
        unpack_singleton: bool = True,
        return_info: bool = False,
    ) -> Databox | tuple[Databox, _Info]:
        r"""
················································································

==Run Kalman filter on a model using time series data==

Executes a Kalman filter on a model, compliant with `KalmanFilterableProtocol`,
using time series observations from the input Databox. This method enables state
estimation and uncertainty quantification in line with the model's dynamics and
the time series data.

    kalman_output = self.kalman_filter(
        input_db,
        span,
        diffuse_scale=None,
        return_=("predict", "update", "smooth", "predict_err", "predict_mse_obs", ),
        return_predict=True,
        return_update=True,
        return_smooth=True,
        return_predict_err=True,
        return_predict_mse_obs=True,
        rescale_variance=False,
        likelihood_contributions=True,
        shocks_from_data=False,
        stds_from_data=False,
        prepend_initial=False,
        append_terminal=False,
        deviation=False,
        check_singularity=False,
        unpack_singleton=True,
        return_info=False,
    )

    kalman_output, info = self.kalman_filter(
        ...
        return_info=True,
    )


### Input arguments ###


???+ input "self"
    The model, compliant with `KalmanFilterableProtocol`, performing the
    Kalman filtering.

???+ input "input_db"
    A Databox containing time series data to be used for filtering.

???+ input "span"
    A date span over which the filtering process is executed based on the
    measurement time series.

???+ input "diffuse_scale"
    A real number or `None`, specifying the scale factor for the diffuse
    initialization. If `None`, the default value is used.

???+ input "return_"
    An iterable of strings indicating which steps' results to return:
    "predict", "update", "smooth".

???+ input "return_predict"
    If `True`, return prediction step results.

???+ input "return_update"
    If `True`, return update step results.

???+ input "return_smooth"
    If `True`, return smoothing step results.

???+ input "rescale_variance"
    If `True`, rescale all variances by the optimal variance scale factor
    estimated using maximum likelihood after the filtering process.

???+ input "likelihood_contributions"
    If `True`, return the contributions of individual periods to the overall
    (negative) log likelihood.

???+ input "shocks_from_data"
    If `True`, use possibly time-varying shock values from the data; these
    values are interpreted as the medians (means) of the shocks. If `False`,
    zeros are used for all shocks.

???+ input "stds_from_data"
    If `True`, use possibly time-varying standard deviation values from the
    data. If `False`, currently assigned constant values are used for the
    standard deviations of all shocks.

???+ input "prepend_initial"
    If `True`, prepend observations to the resulting time series to cover
    initial conditions based on the model's maximum lag. No measurement
    observations are used in these initial time periods (even if some are
    available in the input data).

???+ input "append_terminal"
    If `True`, append observations to the resulting time series to cover
    terminal conditions based on the model's maximum lead. No measurement
    observations are used in these terminal time periods (even if some are
    available in the input data).

???+ input "deviation"
    If `True`, the constant vectors in transition and measurement equations are
    set to zeros, effectively running the Kalman filter as deviations from
    steady state (a balanced-growth path)

???+ input "check_singularity"
    If `True`, check the one-step ahead MSE matrix for the measurement variables
    for singularity, and throw a `SingularMatrixError` exception if the matrix
    is singular.

???+ input "unpack_singleton"
    If `True`, unpack `out_info` into a plain dictionary for models with a
    single variant.

???+ input "return_info"
    If `True`, return additional information about the Kalman filtering process.


### Returns ###


???+ returns "kalman_output"
    A Databox containing some of the following items (depending on the user requests):

    | Attribute         | Type       | Description
    |-------------------|---------------------------------------------------
    | `predict_med`     | `Databox`  | Medians from the prediction step
    | `predict_std`     | `Databox`  | Standard deviations from the prediction step
    | `predict_mse_obs` | `list`     | Mean squared error matrices for the prediction step of the available observations of measurement variables
    | `update_med`      | `Databox`  | Medians from the update step
    | `update_std`      | `Databox`  | Standard deviations from the update step
    | `predict_err`     | `Databox`  | Prediction errors
    | `smooth_med`      | `Databox`  | Medians from the smoothing step
    | `smooth_std`      | `Databox`  | Standard deviations from the smoothing step


???+ returns "out_info"
    A dictionary containing additional information about the filtering process,
    such as log likelihood and variance scale. For models with multiple
    variants, `out_info` is a list of such dictionaries. If
    `unpack_singleton=False`, also `out_info` is a one-element list
    containing the dictionary for singleton models, too.

················································································
        """
        work_db = input_db.shallow()
        #if not shocks_from_data:
        #    shock_names = self.get_names(kind=_quantities.ANY_SHOCK, )
        #    work_db.remove(shock_names, strict_names=False, )
        #if not stds_from_data:
        #    std_names = self.get_names(kind=_quantities.ANY_STD, )
        #    work_db.remove(std_names, strict_names=False, )

        num_variants = self.resolve_num_variants_in_context(num_variants, )

        slatable = self.slatable_for_kalman_filter(
            shocks_from_data=shocks_from_data,
            stds_from_data=stds_from_data,
        )

        input_ds = Dataslate.from_databox_for_slatable(
            slatable, work_db, span,
            num_variants=num_variants,
            prepend_initial=prepend_initial,
            append_terminal=append_terminal,
            clip_data_to_base_span=True,
        )

        frame = Frame(
            start=input_ds.periods[0],
            end=input_ds.periods[-1],
            simulation_end=input_ds.periods[-1],
        )

        needs = Needs(
            return_=return_,
            return_predict=return_predict,
            return_update=return_update,
            return_smooth=return_smooth,
            return_predict_err=return_predict_err,
            return_predict_mse_obs=return_predict_mse_obs,
            rescale_variance=rescale_variance,
            likelihood_contributions=likelihood_contributions,
        )

        squid = Squid.from_squidable(self, )
        qid_to_logly = self.create_qid_to_logly()

        logly_within_y = tuple(
            i for i, qid in enumerate(squid.y_qids, )
            if qid_to_logly.get(qid, False)
        )

        qid_to_name = self.create_qid_to_name()
        y_names = [ qid_to_name[qid] for qid in squid.y_qids ]
        name_to_log_name = {
            n: _quantities.wrap_logly(n, )
            for qid, n in enumerate(input_ds.names, )
            if qid_to_logly.get(qid, False, )
        } if qid_to_logly else {}

        output_store = _OutputStore(
            input_ds=input_ds,
            num_variants=0,
            measurement_names=y_names,
            needs=needs,
            name_to_log_name=name_to_log_name,
        ) if needs.output_store else None

        initialize = _ft.partial(
            _initializers.initialize,
            diffuse_method=diffuse_method,
            diffuse_scale=diffuse_scale,
        )

        #
        # Include the variants even though they are not used because otherwise
        # the number of loops is undetermined (infite iterators)
        main_iter = zip(
            range(num_variants, ),
            self.iter_variants(),
            input_ds.iter_variants(),
        )

        out_info = []

        for vid, self_v, input_ds_v in main_iter:

            variant_header = f"[Variant {vid}]"
            when_singularity = _wd.create_stream(
                when_singularity,
                f"Singularity in prediction MSE matrix in {variant_header}",
            ) if check_singularity else None

            solution_v = self_v._get_singleton_solution(deviation=deviation, )
            data_array = input_ds_v.get_data_variant()

            #
            # Initialize median and MSE of the alpha state vector
            #
            init_cov_u = self_v.get_cov_unanticipated_shocks()
            init_med, init_mse, unknown_init_impact = initialize(solution_v, init_cov_u, )

            #
            # Presimulate the impact of anticipated shocks
            #
            all_v_impact = (
                _simulate_anticipated_shocks(self_v, input_ds_v, frame, )
                if shocks_from_data else None
            )

            #
            # Get values of observables
            #
            y1_array = data_array[squid.y_qids, :].copy()
            if logly_within_y:
                y1_array[logly_within_y, :] = _np.log(y1_array[logly_within_y, :])
            y1_incidence_array = ~_np.isnan(y1_array)
            #
            #
            output_store_v = _OutputStore(
                input_ds=input_ds_v,
                num_variants=1,
                measurement_names=y_names,
                needs=needs,
                name_to_log_name=name_to_log_name,
                transform=solution_v.Ua,
                squid=squid,
                measurement_incidence=y1_incidence_array,
            ) if needs.output_store else None

            generate_period_system = _ft.partial(
                _generate_period_system,
                solution_v=solution_v,
                y1_array=y1_array,
                std_u_array=data_array[squid.std_u_qids, :],
                std_w_array=data_array[squid.std_w_qids, :],
                all_v_impact=all_v_impact,
                initials=(init_med, init_mse, unknown_init_impact, ),
            )

            generate_period_data = _ft.partial(
                _generate_period_data,
                y_array=y1_array,
                u_array=data_array[squid.u_qids, :],
                v_array=data_array[squid.v_qids, :],
                w_array=data_array[squid.w_qids, :],
            )

            store_predict = (
                _ft.partial(_OutputStore.store_predict, self=output_store_v, )
                if needs.return_predict else None
            )

            store_update = (
                _ft.partial(_OutputStore.store_update, self=output_store_v, )
                if needs.return_update else None
            )

            store_smooth = (
                _ft.partial(_OutputStore.store_smooth, self=output_store_v, )
                if needs.return_smooth else None
            )

            cache = predict(
                num_periods=input_ds_v.num_periods,
                generate_period_system=generate_period_system,
                generate_period_data=generate_period_data,
                store_predict=store_predict,
                store_update=store_update,
                store_smooth=store_smooth,
                check_singularity=check_singularity,
                when_singularity=when_singularity,
            )

            if when_singularity is not None:
                when_singularity._raise()

            if cache.needs_estimate_unknown_init:
                estimate_unknown_init(
                    cache=cache,
                )
                correct_for_unknown_init(
                    cache=cache,
                    store_predict=store_predict,
                )

            if needs.return_update:
                update(cache=cache, store_update=store_update, )

            if needs.return_smooth:
                smooth(cache=cache, store_smooth=store_smooth, )

            cache.calculate_likelihood(rescale_variance=needs.rescale_variance, )
            if needs.likelihood_contributions:
                cache.calculate_likelihood_contributions()

            if needs.output_store:
                output_store_v.rescale_stds(cache.var_scale, )

            out_info_v = cache.create_out_info(span, )

            if needs.output_store:
                output_store.extend(output_store_v, )

            out_info.append(out_info_v, )

        out_data = (
            output_store.create_out_data()
            if output_store is not None else None
        )
        #
        if return_info:
            out_info = _has_variants.unpack_singleton(
                out_info, self.is_singleton,
                unpack_singleton=True,
            )
            return out_data, out_info,
        else:
            return out_data

    #]


class Cache:
    """
    """
    #[

    _LOG_2_PI = _np.log(2 * _np.pi)

    _slots_to_input = (
        "num_periods",
    )

    _slots_to_preallocate = (
        "all_y",
        "all_num_obs",
        "all_Fi",
        "all_L",
        "all_a0",
        "all_y0",
        "all_u0",
        "all_v0",
        "all_w0",
        "all_Q0",
        "all_pe",
        "all_Zt_Fi",
        "all_T_G_prev",
        "all_H_cov_w",
        "all_P_cov_u",
        "all_Xi",
        "all_Z",
        "all_G",
        "all_M",
    )

    _other_slots = (
        "needs_estimate_unknown_init",
        "unknown_init_estimate",
        "sum_num_obs",
        "sum_log_det_F",
        "sum_pe_Fi_pe",
        "var_scale",
        "neg_log_likelihood",
        "neg_log_likelihood_contributions",
        "all_log_det_F",
        "all_pe_Fi_pe",
    )

    # __slots__ = _slots_to_input + _slots_to_preallocate + _other_slots

    def __init__(
        self: Self,
        **kwargs,
    ) -> None:
        """
        """
        for n in self._slots_to_input:
            setattr(self, n, kwargs[n], )
        for n in self._slots_to_preallocate:
            setattr(self, n, [None] * self.num_periods, )
        for n in self._other_slots:
            setattr(self, n, None, )

    def calculate_likelihood(self: Self, rescale_variance: bool, ) -> None:
        """
        """
        self.all_pe_Fi_pe = tuple(
            pe @ Fi @ pe if pe is not None and Fi is not None else None
            for pe, Fi, in zip(self.all_pe, self.all_Fi, )
        )
        all_det_Fi = tuple(
            float(_np.linalg.det(Fi)) if Fi is not None else None
            for Fi in self.all_Fi
        )
        self.all_log_det_F = tuple(
            float(-_np.log(det_Fi)) if det_Fi is not None else None
            for det_Fi in all_det_Fi
        )
        self.sum_num_obs = sum(i for i in self.all_num_obs if i is not None)
        self.sum_log_det_F = sum(i for i in self.all_log_det_F if i is not None)
        self.sum_pe_Fi_pe = sum(i for i in self.all_pe_Fi_pe if i is not None)
        self.var_scale = 1
        #
        if rescale_variance:
            self._calculate_variance_scale()
        #
        self.neg_log_likelihood = (
            + self.sum_num_obs*self._LOG_2_PI
            + self.sum_log_det_F
            + self.sum_pe_Fi_pe
        ) / 2;

    def calculate_likelihood_contributions(self: Self, ) -> None:
        """
        """
        self.neg_log_likelihood_contributions = tuple(
            (log_det_F + pe_Fi_pe + num_obs * self._LOG_2_PI)/2 if num_obs else 0
            for log_det_F, pe_Fi_pe, num_obs in zip(
                self.all_log_det_F, self.all_pe_Fi_pe, self.all_num_obs,
            )
            # if num_obs is not None
        )

    def _calculate_variance_scale(self: Self, /, ) -> None:
        """
        """
        if self.sum_num_obs == 0:
            self.var_scale = 1
            self.sum_pe_Fi_pe = 0
            return
        self.var_scale = self.sum_pe_Fi_pe / self.sum_num_obs
        self.sum_log_det_F += self.sum_num_obs * _np.log(self.var_scale)
        self.sum_pe_Fi_pe = self.sum_pe_Fi_pe / self.var_scale

    def create_out_info(self: Self, span: Iterable[Period], ) -> dict[str, Any]:
        """
        """
        out_info = {
            "neg_log_likelihood": float(self.neg_log_likelihood),
            "log_det_F": Series(periods=span, values=self.all_log_det_F, ),
            "var_scale": float(self.var_scale),
            "std_scale": float(_sqrt_positive(self.var_scale)),
        }
        if self.neg_log_likelihood_contributions is not None:
            out_info["neg_log_likelihood_contributions"] = Series(
                periods=span,
                values=self.neg_log_likelihood_contributions,
            )
        return out_info

    #]


def predict(
    num_periods: int,
    generate_period_system: Callable,
    generate_period_data: Callable,
    store_predict: Callable | None = None,
    store_update: Callable | None = None,
    store_smooth: Callable | None = None,
    check_singularity: bool = False,
    when_singularity: _wd.Stream | None = None,
) -> None:
    """
    """
    #[
    cache = Cache(num_periods=num_periods, )
    a1_prev, Q1_prev, Xi_prev = generate_period_system(-1, )
    needs_estimate_unknown_init = Xi_prev is not None
    cache.needs_estimate_unknown_init = needs_estimate_unknown_init
    cache.last_period_of_observations = -1
    create_empty = _ft.partial(_np.empty, shape=(0, 0), dtype=_np.float64, )
    #
    for t in range(num_periods, ):
        #
        T, P, K, Z, H, D, cov_u, cov_w, v_impact, U, *_ = generate_period_system(t, )
        y1, u0, v0, w0, inx_y, *_ = generate_period_data(t, )
        #
        cache.all_y[t] = y1
        cache.all_num_obs[t] = y1.size
        any_y = cache.all_num_obs[t] > 0
        if any_y:
            cache.last_period_of_observations = t
        #
        if t > 0 and store_smooth:
            T_G_prev = T @ G_prev
            cache.all_T_G_prev[t-1] = T_G_prev
            cache.all_L[t-1] = T - T_G_prev @ Z_prev
        #
        # MSE prediction step
        #
        P_cov_u = P @ cov_u
        Q0 = T @ Q1_prev @ T.T + P_cov_u @ P.T
        Q0 = _covariances.symmetrize(Q0)
        H_cov_w = H @ cov_w
        #
        F = (
            Z @ Q0 @ Z.T + H_cov_w @ H.T if any_y
            else create_empty()
        )
        F = _covariances.symmetrize(F, )
        #
        # sing_values = _np.linalg.svd(F, compute_uv=False, )
        # relative_cond = sing_values[-1] / sing_values[0]
        # print(t, F.size, relative_cond, sing_values[0], )
        # TODO: Check if F singularity if requested by user
        inv = _INVERSE_FUNCTION["regular"]
        if when_singularity and any_y:
            inv = _check_singularity(F, t, inx_y, when_singularity, )

        Fi = inv(F) if any_y else create_empty()
        Fi = _covariances.symmetrize(Fi, )
        #
        # Median prediction step
        #
        a0 = T @ a1_prev + K + P @ u0 + (v_impact if v_impact is not None else 0)
        y0 = Z @ a0 + D + H @ w0
        #
        if store_predict:
            store_predict(t=t, a0=a0, y0=y0, u0=u0, v0=v0, w0=w0, Q0=Q0, F=F, cov_u0=cov_u, cov_w0=cov_w, )
        #
        # MSE updating step
        #
        Zt_Fi = Z.T @ Fi
        G = Q0 @ Zt_Fi
        Q1 = Q0 - G @ Z @ Q0
        Q1 = _covariances.symmetrize(Q1)
        if store_update:
            store_update(t=t, Q=Q1, )
        #
        # Median updating step
        #
        pe = y1 - y0
        a1 = a0 + G @ pe
        x1 = U @ a1
        #
        cache.all_a0[t] = a0
        cache.all_y0[t] = y0
        cache.all_pe[t] = pe
        cache.all_Fi[t] = Fi
        cache.all_Z[t] = Z
        #
        if store_smooth:
            cache.all_u0[t] = u0
            cache.all_v0[t] = v0
            cache.all_w0[t] = w0
            cache.all_Q0[t] = Q0
            cache.all_Zt_Fi[t] = Zt_Fi
            cache.all_P_cov_u[t] = P_cov_u
            cache.all_H_cov_w[t] = H_cov_w
        #
        if needs_estimate_unknown_init:
            Xi = ((T - T @ G_prev @ Z_prev) @ Xi_prev) if t > 0 else (T @ Xi_prev)
            cache.all_Xi[t] = Xi
            cache.all_G[t] = G
        #
        a1_prev = a1
        Q1_prev = Q1
        G_prev = G
        Z_prev = Z
        Xi_prev = Xi if needs_estimate_unknown_init else None
    #
    return cache
    #]


def estimate_unknown_init(
    cache: Cache,
    delta_tolerance: float = _DEFAULT_DELTA_TOLERANCE,
) -> None:
    """
    """
    #[
    all_M = tuple(
        (Z @ Xi) if Z is not None and Xi is not None else None
        for Z, Xi in zip(cache.all_Z, cache.all_Xi, )
    )
    all_Mt_Fi = tuple(
        (M.T @ Fi) if M is not None and Fi is not None else None
        for M, Fi in zip(all_M, cache.all_Fi, )
    )
    sum_Mt_Fi_M = sum(
        (Mt_Fi @ M) if Mt_Fi is not None and M is not None else 0
        for Mt_Fi, M in zip(all_Mt_Fi, all_M, )
    )
    sum_Mt_Fi_pe = sum(
        (Mt_Fi @ pe) if Mt_Fi is not None and pe is not None else 0
        for Mt_Fi, pe in zip(all_Mt_Fi, cache.all_pe, )
    )
    #
    sum_Mt_Fi_M = _covariances.symmetrize(sum_Mt_Fi_M)
    sum_Mt_Fi_M[_np.abs(sum_Mt_Fi_M) < delta_tolerance] = 0
    #
    delta, *_ = _np.linalg.lstsq(sum_Mt_Fi_M, sum_Mt_Fi_pe, rcond=None, )
    #
    cache.unknown_init_estimate = delta
    cache.all_M = all_M
    #]


def correct_for_unknown_init(
    cache: Cache,
    store_predict: Callable | None,
) -> None:
    """
    """
    #[
    delta = cache.unknown_init_estimate
    for t in range(cache.num_periods, ):
        cache.all_a0[t] += cache.all_Xi[t] @ delta
        M_delta = cache.all_M[t] @ delta
        cache.all_y0[t] += M_delta
        cache.all_pe[t] -= M_delta
        if store_predict:
            store_predict(t=t, a0=cache.all_a0[t], y0=cache.all_y0[t], )
    #]


def update(
    cache: Cache,
    store_update: Callable | None,
) -> None:
    """
    """
    #[
    for t in range(cache.num_periods, ):
        a1, y1, u1, v1, w1, Q1, *_ = one_step_back(t, cache, )
        if store_update:
            pe = cache.all_pe[t]
            store_update(t=t, xi=a1, y=y1, u=u1, v=v1, w=w1, pe=pe, )
    #]


def smooth(
    cache: Cache,
    store_smooth: Callable | None,
) -> None:
    """
    """
    #[
    N = None
    r = None
    for t in reversed(range(cache.num_periods, )):
        a2, y2, u2, v2, w2, Q2, N, r = one_step_back(t, cache, N, r, )
        if store_smooth:
            store_smooth(t=t, xi=a2, y=y2, u=u2, v=v2, w=w2, Q=Q2, )
    #]


def one_step_back(
    t: int,
    cache: Cache,
    N: _np.ndarray | None = None,
    r: _np.ndarray | None = None,
    /,
) -> tuple:
    """
    """
    #[
    ak = cache.all_a0[t]
    y = cache.all_y[t]
    uk = cache.all_u0[t]
    vk = cache.all_v0[t]
    wk = cache.all_w0[t]
    Qk = cache.all_Q0[t]
    #
    if t <= cache.last_period_of_observations:
        T_G_prev = cache.all_T_G_prev[t]
        P_cov_u = cache.all_P_cov_u[t]
        H_cov_w = cache.all_H_cov_w[t]
        #
        a0 = cache.all_a0[t]
        Q0 = cache.all_Q0[t]
        L = cache.all_L[t]
        Zt_Fi = cache.all_Zt_Fi[t]
        Fi = cache.all_Fi[t]
        Z = cache.all_Z[t]
        pe = cache.all_pe[t]
        #
        Fi_pe = Fi @ pe
        Zt_Fi_pe = Zt_Fi @ pe
        Zt_Fi_Z = Zt_Fi @ Z
        #
        if N is None:
            N = Zt_Fi_Z
        else:
            N = Zt_Fi_Z + L.T @ N @ L
        Qk = Qk - Q0 @ N @ Q0
        Qk = _covariances.symmetrize(Qk)
        #
        if r is None:
            wk = wk + H_cov_w.T @ Fi_pe
            r = Zt_Fi_pe
        else:
            wk = wk + H_cov_w.T @ (Fi_pe - (T_G_prev.T @ r))
            r = Zt_Fi_pe + L.T @ r
        ak = a0 + Q0 @ r
        uk = uk + P_cov_u.T @ r
        vk = vk
    #
    return ak, y, uk, vk, wk, Qk, N, r
    #]


def _generate_period_system(
    t: int,
    #
    solution_v: Solution,
    y1_array: _np.ndarray,
    std_u_array: _np.ndarray,
    std_w_array: _np.ndarray,
    all_v_impact: Sequence[_np.ndarray | None] | None,
    initials: tuple[_np.ndarray, _np.ndarray, int, ],
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    if t == -1:
        return initials
    #
    T = solution_v.Ta
    P = solution_v.Pa
    K = solution_v.Ka
    inx_y = ~_np.isnan(y1_array[:, t], )
    Z = solution_v.Za[inx_y, :]
    H = solution_v.H[inx_y, :]
    D = solution_v.D[inx_y]
    U = solution_v.Ua
    cov_u = _np.diag(std_u_array[:, t]**2, )
    cov_w = _np.diag(std_w_array[:, t]**2, )
    v_impact = all_v_impact[t] if all_v_impact is not None else None
    return T, P, K, Z, H, D, cov_u, cov_w, v_impact, U
    #]


def _generate_period_data(
    t,
    y_array: _np.ndarray,
    u_array: _np.ndarray,
    v_array: _np.ndarray,
    w_array: _np.ndarray,
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    inx_y = ~_np.isnan(y_array[:, t], )
    y = y_array[inx_y, t]
    u = u_array[:, t]
    v = v_array[:, t]
    w = w_array[:, t]
    return y, u, v, w, inx_y.tolist(),
    #]


_INVERSE_FUNCTION = {
    "regular": _np.linalg.inv,
    "singular": _np.linalg.pinv,
}


def _check_singularity(
    F: _np.ndarray,
    t: int,
    boolex_y: list[bool],
    when_singularity: _wd.Stream | None,
) -> Callable | NoReturn:
    """
    """
    rank_F = _np.linalg.matrix_rank(F, )
    if rank_F == F.shape[0]:
        return _INVERSE_FUNCTION["regular"]
    index_y = _np.where(boolex_y, )[0].tolist()
    message = f"Singular prediction MSE matrix in period {t}, rank {rank_F} of {F.shape[0]}"
    # u, s, v, = _np.linalg.svd(F, )
    when_singularity.add(message, )
    return _INVERSE_FUNCTION["singular"]




"""
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp
from numbers import Number

from ..dataslates import Dataslate
from ..fords.solutions import Solution
from ._invariants import Invariant

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self, Iterable
    from ..dates import Period

#]


class System:
    r"""
    """
    #[

    __slots__ = (
        "A",
        "B",
        "c",
        "cov_residuals",
    )

    def __init__(
        self,
        A: _np.ndarray | None = None,
        B: _np.ndarray | None = None,
        c: _np.ndarray | None = None,
        cov_residuals: _np.ndarray | None = None,
    ) -> None:
        r"""
        """
        self.A = A
        self.B = B
        self.c = c
        self.cov_residuals = cov_residuals

    def copy(self, ) -> Self:
        r"""
        """
        new = type(self)()
        for n in self.__slots__:
            attr = getattr(self, n)
            if attr is not None:
                setattr(new, n, attr.copy())
        return new

    @property
    def _can_calculate_dimensions(self, ) -> bool:
        r"""
        """
        return self.A is not None

    @property
    def num_endogenous(self, ) -> int:
        r"""
        """
        if self._can_calculate_dimensions:
            return self.A.shape[0]
        else:
            return None

    @property
    def order(self, ) -> int:
        r"""
        """
        if self._can_calculate_dimensions:
            return self.A.shape[1] // self.num_endogenous
        else:
            return None

    @property
    def has_intercept(self, ) -> bool:
        r"""
        """
        return self.c is not None

    @property
    def num_exogenous(self, ) -> int:
        r"""
        """
        if self._can_calculate_dimensions:
            return self.B.shape[1]
        else:
            return None

    @property
    def num_lagged_endogenous(self, ) -> int:
        r"""
        """
        if self._can_calculate_dimensions:
            return self.num_endogenous * self.order
        else:
            return None

    #]


class Variant:
    """
    """
    #[

    __slots__ = (
        "system",
        "fitted_periods",
        "residual_estimates",

        "_companion_T",
        "_eigenvalues",
        "_max_abs_eigenvalue",
    )

    def __init__(
        self,
        A: _np.ndarray | None = None,
        B: _np.ndarray | None = None,
        c: _np.ndarray | None = None,
        cov_residuals: _np.ndarray | None = None,
        fitted_periods: Iterable[Period] | None = (),
        residual_estimates: _np.ndarray | None = None,
        where_observations: _np.ndarray | None = None,
    ) -> None:
        r"""
        """
        self.system = System(A=A, B=B, c=c, cov_residuals=cov_residuals, )
        self.fitted_periods = tuple(fitted_periods)
        self.residual_estimates = residual_estimates
        #
        self._eigenvalues = None
        self._max_abs_eigenvalue = None
        self._companion_T = None

    def copy(self, ) -> Self:
        r"""
        """
        new = type(self)()
        new.system = self.system.copy()
        new.fitted_periods = self.fitted_periods
        new._eigenvalues = self._eigenvalues
        new._max_abs_eigenvalue = self._max_abs_eigenvalue
        new._companion_T = self._companion_T
        return new

    @property
    def eigenvalues(self, ) -> tuple[Number, ...]:
        r"""
        """
        if self._eigenvalues is None:
            self._populate_eigenvalues()
        return self._eigenvalues

    def _populate_eigenvalues(self, ) -> None:
        r"""
        """
        T = self.companion_T
        if T is None:
            self._eigenvalues = None
            self._max_abs_eigenvalue = None
            return
        eigenvalues = _np.linalg.eigvals(T)
        max_abs_eigenvalue = _np.max(_np.abs(eigenvalues))
        self._eigenvalues = _tuple_from_flat_array(eigenvalues, )
        self._max_abs_eigenvalue = _number_from_numpy(max_abs_eigenvalue, )

    @property
    def max_abs_eigenvalue(self, ) -> Number:
        r"""
        """
        if self._max_abs_eigenvalue is None:
            self._populate_eigenvalues()
        return self._max_abs_eigenvalue

    @property
    def is_stable(self, ) -> bool:
        r"""
        """
        return (
            self.max_abs_eigenvalue < 1
            if self.max_abs_eigenvalue is not None
            else None
        )

    @property
    def companion_T(self, ) -> _np.ndarray:
        r"""
        """
        if self._companion_T is None:
            self._populate_companion_T()
        return self._companion_T

    def _populate_companion_T(self, ) -> None:
        r"""
        """
        if self.system.A is None:
            self._companion_T = None
            return
        num_endogenous = self.system.num_endogenous
        order = self.system.order
        num_extra = num_endogenous * (order - 1)
        dynamic_identity = _np.eye(num_extra, num_endogenous * order, )
        self._companion_T = _np.concatenate((
            self.system.A,
            dynamic_identity,
        ))

    def _get_companion_P(self, ) -> _np.ndarray:
        r"""
        """
        num_endogenous = self.system.num_endogenous
        num_lagged_endogenous = self.system.num_lagged_endogenous
        return _np.eye(num_lagged_endogenous, num_endogenous, dtype=float, )

    def _get_companion_K(self, ) -> _np.ndarray:
        r"""
        """
        num_endogenous = self.system.num_endogenous
        num_lagged_endogenous = self.system.num_lagged_endogenous
        if self.system.c is None:
            return _np.zeros((num_lagged_endogenous, ), dtype=float, )
        else:
            return _np.concatenate((
                self.system.c,
                _np.zeros((num_lagged_endogenous - num_endogenous, ), dtype=float, ),
            ))

    def _get_companion_sigma(self, ) -> tuple[_np.ndarray, _np.ndarray, ]:
        r"""
        """
        num_endogenous = self.system.num_endogenous
        order = self.system.order
        num_extra = num_endogenous * (order - 1)
        extra_zeros = _np.zeros((num_extra, num_extra), dtype=self.system.cov_residuals.dtype, )
        return _sp.linalg.block_diag(
            self.system.cov_residuals, extra_zeros,
        )

    def get_acov(
        self,
        up_to_order: int = 0,
    ) -> _np.ndarray:
        r"""
        """
        num_endogenous = self.system.num_endogenous
        companion_T = self.companion_T
        companion_sigma = self._get_companion_sigma()
        Omega = _sp.linalg.solve_discrete_lyapunov(companion_T, companion_sigma, )
        acov = []
        acov.append(Omega[:num_endogenous, :num_endogenous])
        for i in range(up_to_order):
            Omega = companion_T @ Omega
            acov.append(Omega[:num_endogenous, :num_endogenous])
        return tuple(acov)

    def get_mean(self, ) -> _np.ndarray:
        r"""
        """
        num_endogenous = self.system.num_endogenous
        order = self.system.order
        A = self.system.A
        c = self.system.c
        if c is None or _np.all(c == 0):
            return _np.zeros((num_endogenous, ), dtype=A.dtype, )
        new_shape = (num_endogenous, num_endogenous, order, )
        A = A.reshape(new_shape, order="F", )
        A = A.sum(axis=2, )
        I = _np.eye(num_endogenous, )
        return _np.linalg.solve(I - A, c)

    def _get_companion_solution(
        self,
        deviation: bool = False,
    ) -> Solution:
        r"""
        """
        solution = Solution(
            T=self.companion_T,
            P=self._get_companion_P(),
            K=self._get_companion_K(),
        )
        if deviation:
            solution = Solution.deviation_solution(solution, )
        return solution

    #]


def _tuple_from_flat_array(array: _np.ndarray, ) -> tuple[Number, ...]:
    r"""
    """
    return tuple(
        _number_from_numpy(i)
        for i in array.flatten()
    )


def _number_from_numpy(x) -> Number:
    r"""
    """
    return float(_np.real(x)) if _np.isreal(x) else complex(x)





import sys
sys.path.append("../..")
import json
import os
import numpy as np_
import scipy as sp_
import warnings

from irispie import *


def glogp(*args):
    y0, sigma, nu, low, high = args
    low = y0 + low
    high = y0 + high;
    mu = sigma * np_.log( ((y0-low)/(high-low)) ** (-1/np_.exp(nu)) - 1 );
    return mu, sigma, nu, low, high


def glogc(x, *args):
    mu, sigma, nu, low, high = glogp(*args)
    nu1 = np_.exp(nu)
    z = (x - mu) / sigma
    y = (1 + np_.exp(-z)) ** (-nu1)
    return low + (high - low) * y


def glogd(x, *args):
    mu, sigma, nu, low, high = glogp(*args)
    nu1 = np_.exp(nu)
    sigma1 = 1 / sigma
    z = (x - mu) * sigma1
    dy = nu1 * sigma1 * (1 + exp(-z))**(-nu1-1) * np_.exp(-z);
    return (high - low) * dy


with open(os.path.join("parameters", "parameters.json"), "r") as fid:
    p = json.load(fid)

with open(os.path.join("parameters", "steady.json"), "r") as fid:
    s = json.load(fid)


source_files = [
  "model-source/macroWorld.model",
  "model-source/macroLocal.model",
  "model-source/macroAssets.model",

  "model-source/connectCreditCreation.model",
  "model-source/connectCreditRisk.model",
  "model-source/connectInterestRates.model",

  "model-source/bankLoanPerformance.model",
  "model-source/bankProvisioning.model",
  "model-source/bankCapital.model",
]


context = {
    "segments": ["hh"],
    "glogc": glogc,
    "glogd": glogd,
}


m = Model.from_file(
    source_files,
    context=context,
    linear=False,
    flat=False,
)

m.assign(**p)
# m.assign(**s)

m.assign(
    py = 1,
    y = 1,
)

q = m._get_quantities(kind=TRANSITION_VARIABLE | MEASUREMENT_VARIABLE)
q = filter_quantities_by_name(q, exclude_names=["y", "py"])

s = m.create_steady_evaluator(quantities=q, print_iter=True)

r1 = sp_.optimize.root(
    s.eval_with_jacobian,
    s.initial_guess,
    method="lm",
    jac=True,
)





import irispie as ir


db = ir.Databox()
db["a"] = 1
db["b"] = 2
db["c"] = "abc"
db["A"] = None
db["B"] = [10, 20, 30, ]


def test_has_single_name():
    assert db.has("a")
    assert db.has("b")
    assert db.has("c")
    assert db.has("A")
    assert db.has("B")
    assert not db.has("d")
    assert not db.has("C")
    assert not db.has("D")


def test_has_tuple_of_names():
    assert db.has(("a", "b"))
    assert db.has(("b", "c"))
    assert db.has(("c", "A"))
    assert db.has(("A", "B"))
    assert not db.has(("a", "C"))
    assert not db.has(("c", "D"))
    assert not db.has(("D", "C"))


def test_shallow_copy():
    sh = db.shallow(source_names=("B", ))
    sh["B"].append(40)
    assert db["B"] == [10, 20, 30, 40, ]


def test_remove_single():
    sh = db.shallow()
    sh.remove("a")
    assert set(sh.keys()) == set(db.keys()) - {"a", }


def test_remove_multiple():
    sh = db.shallow()
    sh.remove(("b", "c"))
    assert set(sh.keys()) == set(db.keys()) - {"b", "c", }


def test_keep_single():
    sh = db.shallow()
    sh.keep("a")
    assert set(sh.keys()) == {"a", }


def test_keep_multiple():
    sh = db.shallow()
    sh.keep(("b", "c"))
    assert set(sh.keys()) == {"b", "c", }


def test_get_names_with_filter():
    sh = db.shallow()
    names = sh.get_names(filter=lambda n: n[0].isupper())
    assert set(names) == {"A", "B", }





"""
Meta plans for steady-state calculations
"""


#[

from __future__ import annotations

from typing import (TYPE_CHECKING, Protocol, )
from numbers import (Real, )
import textwrap as _tw
import documark as _dm

from ..conveniences import copies as _copies
from ..series.main import (Series, )
from . import _registers as _registers
from . import _pretty as _pretty

if TYPE_CHECKING:
    from typing import (Iterable, )

#]


__all__ = (
    "SteadyPlan",
)


if TYPE_CHECKING:
    _InputNames = Iterable[str] | str | EllipsisType


class SimulationPlannableProtocol(Protocol, ):
    """
    """
    #[

    steady_can_be_exogenized: Iterable[str] | None
    steady_can_be_endogenized: Iterable[str] | None
    steady_can_be_fixed_level: Iterable[str] | None
    steady_can_be_fixed_change: Iterable[str] | None

    #]



#[
@_dm.reference(
    path=("structural_models", "steady_plans.md", ),
    categories={
        "constructor": "Creating new steady plans",
        "exogenizing": "Exogenizing and endogenizing steady-state values",
        "fixing": "Fixing steady-state values",
        "information": "Getting information about steady plans",
    },
)
#]
class SteadyPlan(
    _registers.Mixin,
    _pretty.Mixin,
    _copies.Mixin,
):
    """
    """
    #[

    _TABLE_FIELDS = ("NAME", "REGISTER", "VALUE", )

    _registers = (
        "exogenized",
        "endogenized",
        "fixed_level",
        "fixed_change",
    )

    __slots__ = (
        ()
        + tuple(f"_can_be_{r}" for r in _registers)
        + tuple(f"_{r}_register" for r in _registers)
    )

    @_dm.reference(
        category="constructor",
        call_name="SteadyPlan",
    )
    def __init__(
        self,
        model,
        **kwargs,
    ) -> None:
        r"""
················································································

==Create new steady plan object==

................................................................................
        """
        plannable = model.get_steady_plannable(**kwargs, )
        def default_value(*args, **kwargs, ):
            return False
        self._initialize_registers(plannable, default_value, )

    for method_name, register_name in (
        ("exogenize", "exogenized", ),
        ("endogenize", "endogenized", ),
        ("fix_level", "fixed_level", ),
        ("fix_change", "fixed_change", ),
    ):
        exec(_tw.dedent(f"""
            def {method_name}(self, names: _InputNames, ) -> None:
                self._write_to_register("{register_name}", names, True, )

            def un{method_name}(self, names: _InputNames, ) -> None:
                self._write_to_register("{register_name}", names, False, )

            def get_{register_name}_names(self, /, ) -> tuple[str, ...]:
                return self._get_names_from_register("{register_name}", )
        """))

    # TODO: pluralize as default
    fix_levels = fix_level
    fix_changes = fix_change

    @_dm.reference(category="fixing", )
    def fix(self, *args, **kwargs, ) -> None:
        r"""
................................................................................

==Fix steady-state values==

................................................................................
        """
        self.fix_level(*args, **kwargs, )
        if self._fixed_change_register:
            self.fix_change(*args, **kwargs, )

    @_dm.reference(category="fixing", )
    def unfix(self, *args, **kwargs, ) -> None:
        r"""
................................................................................

==Unfix steady-state values==

................................................................................
        """
        self.unfix_level(*args, **kwargs, )
        if self._fixed_change_register:
            self.unfix_change(*args, **kwargs, )

    def swap(self, *args, ) -> None:
        """
        """
        for a in args:
            self.exogenize(a[0], )
            self.endogenize(a[1], )

    def unswap(self, *args, ) -> None:
        """
        """
        for a in args:
            self.unexogenize(a[0], )
            self.unendogenize(a[1], )

    @property
    def is_empty(self, /, ) -> bool:
        """
        True if there are no exogenized, endogenized or fixed names in the plan
        """
        return not any( self.any_in_register(r, ) for r in self._registers )

    def any_in_register(self, register_name: str, /, ) -> bool:
        """
        """
        return any(self.get_register_by_name(register_name, ).values(), )

    def _write_to_register(
        self,
        register_name: str,
        names: _InputNames,
        new_status: bool,
    ) -> None:
        """
        """
        register = self.get_register_by_name(register_name, )
        names = self._resolve_register_names(register, names, )
        for n in names:
            register[n] = new_status

    def _get_names_from_register(
        self,
        register_name: str,
    ) -> tuple[str, ...]:
        """
        """
        register = self.get_register_by_name(register_name, )
        return tuple(
            name for name, value in register.items()
            if value
        )

    def _add_register_to_table(
        self,
        table,
        register: dict,
        action: str,
        model = None,
        *args, **kwargs,
    ) -> None:
        """
        """
        #
        missing_str = Series._missing_str
        level_values = model.get_steady_levels() if model else {}
        change_values = model.get_steady_changes() if model else {}
        parameter_values = model.get_parameters() if model else {}
        def _print_value(value: Real | None, ) -> str:
            return f"{value:g}" if value is not None else missing_str
        #
        def _get_exogenized_value(name, ):
            return (
                "("
                + _print_value(level_values.get(name, None))
                + ", "
                + _print_value(change_values.get(name, None))
                + ")"
            )
        #
        def _get_endogenized_value(name, ):
            return _print_value(parameter_values.get(name, None), )
        #
        def _get_fixed_level_value(name, ):
            return _print_value(level_values.get(name, None), )
        #
        def _get_fixed_change_value(name, ):
            return _print_value(level_values.get(name, None), )
        #
        GET_VALUE_FUNC = {
            "exogenized": _get_exogenized_value,
            "endogenized": _get_endogenized_value,
            "fixed_level": _get_fixed_level_value,
            "fixed_change": _get_fixed_change_value,
        }
        #
        def _get_value(name, action, ):
            return GET_VALUE_FUNC[action](name, )
        #
        all_rows = (
            (k, action, _get_value(k, action, ), )
            for k, v in register.items() if v
        )
        all_rows = sorted(all_rows, key=lambda row: (row[0], row[1], ), )
        for row in all_rows:
            table.add_row(row, )

    #]






import sys
sys.path.append("../..")
from irispie import *
from irispie.parsers import preparser as pr_
import json
import numpy as np_


import nested
import model_functions as mf


def glogp(*args):
    y0, sigma, nu, low, high = args
    low = y0 + low
    high = y0 + high;
    mu = sigma * np_.log( ((y0-low)/(high-low)) ** (-1/np_.exp(nu)) - 1 );
    return mu, sigma, nu, low, high


def glogc(x, *args):
    mu, sigma, nu, low, high = glogp(*args)
    nu1 = np_.exp(nu)
    z = (x - mu) / sigma
    y = (1 + np_.exp(-z)) ** (-nu1)
    return low + (high - low) * y


def glogd(x, *args):
    mu, sigma, nu, low, high = glogp(*args)
    nu1 = np_.exp(nu)
    sigma1 = 1 / sigma
    z = (x - mu) * sigma1
    dy = nu1 * sigma1 * (1 + exp(-z))**(-nu1-1) * np_.exp(-z);
    return (high - low) * dy


with open("parameters.json", "r") as fid:
    p = Databank.from_dict(json.load(fid))


source_string = r"""
    !transition-variables
        q, z_unc, z, a, b

    !parameters
        ss_q, c2, c3, c4, c5

    !transition-equations
        q = glogc(-(z-1), ss_q, c2, c3, c4, c5);
        z_unc = 0.8*z_unc[-1] + (1-0.8)*1;
        z = maximum(z, 0);
        a = log(b);
        b = 1;
"""


context = {
    "segments": ["hh"],
    "glogc": glogc,
}


m = Model.from_string(
    source_string,
    context=context,
    linear=False,
    flat=True,
)

m.assign(
    ss_q = 0.02,
    q = 0.02,
    z_unc = 1,
    z = 1,
    b = 1,
    a = 0,
    c2 = p.c2_q_hh,
    c3 = p.c3_q_hh,
    c4 = p.c4_q_hh,
    c5 = p.c5_q_hh,
)

# m.alter_num_variants(3)

chk = m.check_steady(details=True, when_fails="error")

s = m.systemize()





import json as _js
import plotly as _pl
import irispie as _ir


m = _ir.Model.from_file(
    ["model-source/vanilla-rbc.model", "model-source/parameters.model"],
)

n = _ir.Model.from_file(
    ["model-source/vanilla-rbc-stationarized.model", "model-source/parameters.model"],
    flat=True,
)

n.postprocessor = _ir.Sequential.from_file("model-source/vanilla-rbc-stationarized-postprocessor.model", )

parameters = dict(
    alpha = 1.02**(1/4),
    beta = 0.95**(1/4),
    gamma = 0.40,
    delta = 0.05,
    rho = 0.8,
    std_shock_a = 0.9,
    std_shock_c = 0.9,
)

with open("parameters.json", "wt+", ) as fid:
    _js.dump(parameters, fid, indent=4, )

m.assign(**parameters, )
n.assign(**parameters, )

m.assign(a = 1, k = 20, )
n.assign(kk = 20, )

fix_level = ("a", )

info = m.steady(
    fix_level=fix_level,
    flat=False,
    iter_printer_settings={"every": 5, },
)

n.steady(
    iter_printer_settings={"every": 5, },
)

chk, info = m.check_steady(when_fails="warning", )
chk, info = n.check_steady(when_fails="warning", )

m.solve()
n.solve()

start_sim = _ir.qq(2020,1)
end_sim = _ir.qq(2040,4)
sim_range = start_sim >> end_sim


dm = _ir.Databox.steady(m, sim_range, deviation=True, )
dm0 = dm.copy()
dm0["shock_a"][start_sim] = 0.1
sm0, *_ = m.simulate(dm0, sim_range, deviation=True, )

dn = _ir.Databox.steady(n, sim_range, deviation=True, )
dn0 = dn.copy()
dn0["shock_a"][start_sim] = 0.1
sn0, *_ = n.simulate(dn0, sim_range, deviation=True, )

dm = _ir.Databox.steady(m, sim_range, )
dm1 = dm.copy()
dm1["shock_a"][start_sim] = 0.1
sm1, *_ = m.simulate(dm1, sim_range, )

dn = _ir.Databox.steady(n, sim_range, )
dn1 = dn.copy()
dn1["shock_a"][start_sim] = 0.1
sn1, *_ = n.simulate(dn1, sim_range, )

sn1['a'] = _ir.Series()
sn1['a'][start_sim-1] = sm1['a'][start_sim-1]
p = _ir.PlanSimulate(n.postprocessor, start_sim-1>>end_sim, )
p.exogenize(start_sim-1, "a", when_data=True, )
sn1, *_ = n.postprocess(sn1, start_sim-1>>end_sim, target_databox=sn1, plan=p, when_nonfinite="silent", )


chart_names = ("c", "i", "r", "a", )
descriptions = m.get_descriptions()
chart_titles = tuple(descriptions[n] for n in chart_names)


fig = _pl.subplots.make_subplots(
    rows=2, cols=2, shared_xaxes=True,
    subplot_titles=chart_titles,
)

for i, n in enumerate(chart_names):
    (sm1[n] | sn1[n]).plot(
        range=start_sim-1 >> end_sim,
        figure=fig,
        legend=["Full level model", "Stationarized model", ] if i == 0 else None,
        subplot=i,
        traces=({"mode": "lines+markers"}, {"mode": "lines"}, ),
    )

fig.show()






import irispie as ir
import pytest
import os


_THIS_DIR = os.path.dirname(os.path.abspath(__file__))
_DATA_FILE_PATH = os.path.join(_THIS_DIR, "fred_data.csv")

db = ir.Databox.from_csv_file(
    _DATA_FILE_PATH,
    period_from_string=ir.Period.from_iso_string,
    description_row=True,
)

db["y"] = 100 * ir.log(db["GDPC"])
db["cpi"] = 100 * ir.log(db["CPI"])
db["ad_y"] = 4 * ir.diff(db["y"])
db["ad_cpi"] = 4 * ir.diff(db["cpi"])
db["rs"] = db["TB3M"]

db["y_tnd"], db["y_gap"], = ir.hpf(db["y"], )
db["ad2_y_tnd"] = 4 * ir.diff(ir.diff(db["y_tnd"]))

estim_start = ir.qq(1950, 1) + 50
estim_end = ir.qq(2022, 4)
estim_span = estim_start >> estim_end
num_estim_periods = len(estim_span)
sim_span = estim_end+1 >> estim_end+8

db["covid_dummy"] = ir.Series(periods=estim_span, values=0, )
db["covid_dummy"][ir.qq(2020,2)] = 1

endogenous_names = ["ad2_y_tnd", "y_gap", "cpi", "rs"]
num_resamples = 1_000


def test_plain():
    v0 = ir.RedVAR(endogenous_names, order=2, )
    est_db0 = v0.estimate(db, estim_span, omit_missing=True, )
    acov0 = v0.get_acov(up_to_order=2, )
    acorr0 = v0.get_acorr(acov=acov0, )
    mean0 = v0.get_mean()
    db["covid_dummy"][sim_span] = 0
    sim_db0 = v0.simulate(db, estim_end+1 >> estim_end+8, prepend_input=False, )
    #
    rdb = v0.resample(
        est_db0,
        estim_span,
        "wild_bootstrap",
        num_variants=num_resamples,
    )
    #
    rv0 = v0.copy()
    #
    rv0.estimate(
        rdb,
        estim_span,
        omit_missing=True,
        num_variants=num_resamples,
    )
    #
    rc0 = rv0.get_acov(up_to_order=1, )
    rr0 = rv0.get_acorr(acov=rc0, )
    #
    rsim_db0 = rv0.simulate(
        db,
        estim_end+1 >> estim_end+8,
        prepend_input=True,
    )


d1 = ir.MinnesotaPriorObs(rho=0, mu2=num_estim_periods, )
d2 = ir.MeanPriorObs(mu2=num_estim_periods, mean=[0,0,0,0], )


def test_minnesota():
    v = ir.RedVAR(endogenous_names, order=2, )
    est_db = v.estimate(
        db, estim_span,
        omit_missing=True,
        prior_obs=d1,
        target_db=db,
    )
    mean = v.get_mean()


def test_mean():
    v = ir.RedVAR(endogenous_names, order=2, )
    est_db = v.estimate(
        db, estim_span,
        omit_missing=True,
        prior_obs=d2,
        target_db=db,
    )
    mean = v.get_mean()


def test_combined():
    v = ir.RedVAR(endogenous_names, order=2, )
    est_db = v.estimate(
        db, estim_span,
        omit_missing=True,
        prior_obs=(d1, d2),
        target_db=db,
    )
    mean = v.get_mean()





import sys
sys.path.append('../..')
from irispie import *
from irispie.parsers import preparser as pp
import scipy as sp_

m = Model.from_file("spbc.model", flat=True, )

m.assign(
    alpha = 1,
    beta = 0.985**(1/4),
    beta0 = 0,
    beta1 = 0.8,
    gamma = 0.60,
    delta = 0.03,
    pi = 1,
    eta = 6,
    k = 10,
    psi = 0.25,

    chi = 0.85,
    xiw = 60,
    xip = 300,
    rhoa = 0.90,
    rhoterm20 = 0.80,

    rhor = 0.85,
    kappap = 4,
    kappan = 0,

    Short_ = 0,
    Long_ = 0,
    Infl_ = 0,
    Growth_ = 0,
    Wage_ = 0,

    A = 1,
    P = 1,
    Pk = 5,
)


select_index = lambda q: [ 
    i for i, j in enumerate(q) 
    if j.human not in ["A", "P", "Short", "Infl", "Long", "Growth", "Wage"]
]

q1 = m._get_quantities(kind=TRANSITION_VARIABLE | MEASUREMENT_VARIABLE)
s1 = m.create_steady_evaluator(quantities=q1)
ix1 = select_index(q1)

q2 = m._get_quantities(kind=TRANSITION_VARIABLE | MEASUREMENT_VARIABLE)
q2 = filter_quantities_by_name(q2, exclude_names=["A", "P"])
s2 = m.create_steady_evaluator(quantities=q2, print_iter=True)
ix2 = select_index(q2)

r1 = sp_.optimize.root(
    s1.eval_with_jacobian,
    s1.initial_guess,
    method="lm",
    jac=True,
)

r2 = sp_.optimize.root(
    s2.eval_with_jacobian,
    s2.initial_guess,
    method="lm",
    jac=True,
)

s3 = m.create_steady_evaluator(quantities=q2, print_iter=False, flat=False)
s3.update()
j3 = s3._jacobian.eval(s3._x, None)




"""
First-order systems, solutions, and estimators
"""

from .solutions import *
from .solutions import __all__ as solutions_all

from .covariances import *
from .covariances import __all__ as covariances_all

from .least_squares import *
from .least_squares import __all__ as least_squares_all

__all__ = (
    *solutions_all,
    *covariances_all,
    *least_squares_all,
)




r"""
Implement SlatableProtocol for RedVAR models
"""


#[

from __future__ import annotations

import warnings as _wa

from ..series import Series
from ..dataslates import Slatable
from .. import quantities as _quantities

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from numbers import Real
    from .red_vars import RedVAR

#]


_DEFAULT_RESIDUAL_VALUE = 0.0


class Inlay:
    r"""
    """
    #[

    def slatable_for_estimate(self, ) -> Slatable:
        r"""
        """
        slatable = _slatable_for_estimate_or_simulate(self, )
        residual_name_to_value = _create_residual_name_to_value(self, )
        slatable.overwrites.update(residual_name_to_value, )
        return slatable

    def slatable_for_simulate(self, residuals_from_data: bool, ) -> Slatable:
        r"""
        """
        slatable = _slatable_for_estimate_or_simulate(self, )
        residual_name_to_value = _create_residual_name_to_value(self, )
        if residuals_from_data:
            slatable.fallbacks.update(residual_name_to_value, )
        else:
            slatable.overwrites.update(residual_name_to_value, )
        return slatable

    #]


def _slatable_for_estimate_or_simulate(model: RedVAR, ) -> Slatable:
    r"""
    Create slatable base for estimate or simulate contexts
    """
    #{
    slatable = Slatable()
    #
    slatable.max_lag = -model.order
    #
    slatable.databox_names = model.get_names()
    slatable.output_names = slatable.databox_names
    #
    name_to_description = model.create_name_to_description()
    slatable.descriptions = tuple(
        name_to_description.get(name, "", )
        for name in slatable.databox_names
    )
    #
    # Fallbacks and overwrites
    slatable.fallbacks = {}
    slatable.overwrites = {}
    #
    # Databox validation - all input data must be time series
    validator = (
        lambda x: isinstance(x, Series),
        "Input data for this variable is not a time series",
    )
    slatable.databox_validators = {
        name: validator
        for name in slatable.databox_names
    }
    #
    return slatable
    #]


def _create_residual_name_to_value(model, ) -> dict[str, Real]:
    r"""
    Create dict with default residual values for all residuals in the model
    """
    #[
    residual_names = model.get_names(kind=_quantities.UNANTICIPATED_SHOCK, )
    return {
        name: _DEFAULT_RESIDUAL_VALUE
        for name in residual_names
    }
    #]




r"""
Implement SlatableProtocol for Sequential models
"""


#[

from __future__ import annotations

from ..dataslates import Slatable

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from .main import Sequential

#]


_DEFAULT_RESIDUAL_VALUE = 0.0


class Inlay:
    r"""
    """
    #[

    def slatable_for_simulate(
        self,
        parameters_from_data: bool,
        shocks_from_data: bool,
    ) -> Slatable:
        """
        """
        slatable = Slatable()
        #
        slatable.max_lag = self.max_lag
        slatable.max_lead = self.max_lead
        #
        slatable.databox_names = self.all_names
        slatable.descriptions = None
        slatable.databox_validators = None
        #
        slatable.fallbacks = {}
        slatable.overwrites = {}
        #
        parameter_name_to_value = self.get_parameters(unpack_singleton=True, )
        if parameters_from_data:
            slatable.fallbacks.update(parameter_name_to_value, )
        else:
            slatable.overwrites.update(parameter_name_to_value, )
        #
        residual_name_to_value = {
            name: _DEFAULT_RESIDUAL_VALUE
            for name in self.residual_names
        }
        if shocks_from_data:
            slatable.fallbacks.update(residual_name_to_value, )
        else:
            slatable.overwrites.update(residual_name_to_value, )
        #
        slatable.qid_to_logly = {}
        #
        slatable.output_names = (
            self.lhs_names
            + self.rhs_only_names
            + self.residual_names
        )
        if parameters_from_data:
            slatable.output_names += self.parameter_names
        #
        return slatable

    #]




r"""
Implement SlatableProtocol for Simultaneous models
"""


#[

from __future__ import annotations

from ..series.main import Series
from .. import quantities as _quantities
from ..dataslates import Slatable

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from numbers import Real

#]


_DEFAULT_SHOCK_VALUE = 0.0


_OUTPUT_KIND_FOR_SIMULATE = (
    _quantities.ANY_VARIABLE
    | _quantities.ANY_SHOCK
)


_OUTPUT_KIND_FOR_KALMAN_FILTER = (
    _quantities.ANY_VARIABLE
    | _quantities.ANY_SHOCK
    | _quantities.UNANTICIPATED_STD
    | _quantities.MEASUREMENT_STD
)


class Inlay:
    r"""
    """
    #[

    def slatable_for_simulate(self, **kwargs, ) -> Slatable:
        r"""
        """
        slatable = _slatable_for_simulate_or_kalman_filter(self, **kwargs, )
        slatable.output_names = self.get_names(kind=_OUTPUT_KIND_FOR_SIMULATE, )
        return slatable

    def slatable_for_kalman_filter(self, **kwargs, ) -> Slatable:
        r"""
        """
        slatable = _slatable_for_simulate_or_kalman_filter(self, parameters_from_data=False, **kwargs, )
        slatable.output_names = self.get_names(kind=_OUTPUT_KIND_FOR_KALMAN_FILTER, )
        return slatable

    def slatables_for_vary_stds(self, ) -> tuple[Slatable, Slatable]:
        r"""
        """
        std_slatable = _slatable_for_vary_stds(self, )
        std_name_to_value = self.get_stds(unpack_singleton=False, )
        std_slatable.fallbacks = std_name_to_value
        #
        multiplier_slatable = _slatable_for_vary_stds(self, )
        multiplier_name_to_value = { name: 1 for name in std_name_to_value.keys() }
        multiplier_slatable.fallbacks = multiplier_name_to_value
        #
        return std_slatable, multiplier_slatable,

    #]


def _slatable_for_simulate_or_kalman_filter(
    model,
    parameters_from_data: bool,
    shocks_from_data: bool,
    stds_from_data: bool,
) -> Slatable:
    r"""
    """
    #[
    slatable = Slatable()
    #
    slatable.max_lag = model.max_lag
    slatable.max_lead = model.max_lead
    #
    qid_to_name = model.create_qid_to_name()
    slatable.databox_names = tuple(
        qid_to_name[qid]
        for qid in sorted(qid_to_name)
    )
    #
    name_to_description = model.create_name_to_description()
    slatable.descriptions = tuple(
        name_to_description.get(name, "", )
        for name in slatable.databox_names
    )
    #
    # Databox validation - all variables must be time series
    variable_names = model.get_names(kind=_quantities.ANY_VARIABLE, )
    validator = (
        lambda x: isinstance(x, Series),
        "Input data for this variable is not a time series",
    )
    slatable.databox_validators = {
        name: validator
        for name in variable_names
    }
    #
    # Fallbacks and overwrites
    slatable.fallbacks = {}
    slatable.overwrites = {}
    #
    parameter_name_to_value = model.get_parameters(unpack_singleton=True, )
    if parameters_from_data:
        slatable.fallbacks.update(parameter_name_to_value, )
    else:
        slatable.overwrites.update(parameter_name_to_value, )
    #
    shock_names = model.get_names(kind=_quantities.ANY_SHOCK, )
    shock_name_to_value = {
        name: [_DEFAULT_SHOCK_VALUE, ]*model.num_variants
        for name in shock_names
    }
    if shocks_from_data:
        slatable.fallbacks.update(shock_name_to_value, )
    else:
        slatable.overwrites.update(shock_name_to_value, )
    #
    std_name_to_value = model.get_stds(unpack_singleton=False, )
    if stds_from_data:
        slatable.fallbacks.update(std_name_to_value, )
    else:
        slatable.overwrites.update(std_name_to_value, )
    #
    slatable.qid_to_logly = model.create_qid_to_logly()
    #
    return slatable
    #]


def _slatable_for_vary_stds(model, ) -> Slatable:
    r"""
    """
    #[
    slatable = Slatable()
    std_qids = _quantities.generate_qids_by_kind(
        model._invariant.quantities,
        kind=_quantities.ANY_STD,
    )
    qid_to_name = model.create_qid_to_name()
    slatable.databox_names = tuple(
        qid_to_name[qid]
        for qid in sorted(std_qids)
    )
    slatable.output_names = slatable.databox_names
    return slatable
    #]





import irispie as ir

model = """
!variables
    x, y, z

!shocks
    shk_x, shk_y, shk_z

!equations
    x = 0.8*x[-1] + (1-0.8)*1.5 + shk_x;
    y = x + shk_y;
    log(z) = 0.8*log(z[-1]) + (1-0.8)*log(3) + shk_z;

"""

m = ir.Model.from_string(model, flat=True, )


m.steady(optim_settings={"factor": 0.1}, )
m.check_steady()

print(m.get_steady(round=4, ), )

a = m.create_steady_array(num_columns=5, )

print(a)





import model as md
import pickle as pk
# import numpy as np
import model_functions as mf

# def cdf(a, b):
    # return np.log(a) + np.log(b)

def main():

    context = {
        "cdf": mf.cdf,
    }

    m = md.Model(context)
    print(m.eval(10, 20))

    with open("model.pkl", "wb") as f:
        pk.dump(m, f)

    return m

if __name__ == "__main__":
    m = main()





import irispie as ir
import random as rn
import functools as ft


start = ir.qq(2020,1)
end = ir.qq(2025,4)

start_missing = ir.qq(2021,4)
end_missing = ir.qq(2022,3)

x = ir.Series(periods=start>>end, func=ft.partial(rn.uniform, 1, 2), )

x_missing = x.copy()
x_missing[start_missing>>end_missing] = None


def test_fill_previous():
    x_filled = ir.fill_missing(x_missing, method="previous", )
    assert x_filled.any_missing(start>>end) == False


def test_fill_next():
    x_filled = ir.fill_missing(x_missing, method="next", )
    assert x_filled.any_missing(start>>end) == False


def test_fill_nearest():
    x_filled = ir.fill_missing(x_missing, method="nearest", )
    assert x_filled.any_missing(start>>end) == False


def test_fill_constant():
    x_filled = ir.fill_missing(x_missing, method="constant", method_args=100, )
    assert x_filled.any_missing(start>>end) == False


def test_fill_log_linear():
    x_filled = ir.fill_missing(x_missing, method="log_linear", )
    assert x_filled.any_missing(start>>end) == False





"""
Vector autoregression (VAR) module
"""

from .red_vars import *
from .red_vars import __all__ as red_vars_all

from .struct_vars import *
from .struct_vars import __all__ as struct_vars_all

from .prior_obs import *
from .prior_obs import __all__ as prior_obs_all

__all__ = (
    *red_vars_all,
    *struct_vars_all,
    *prior_obs_all,
)





import irispie as ir

source = """

!equations{ :abc :_D }

    !for ?x = 0, 1, 2 !do
        !if aaa == ?x !then
            x = 1;
            !if aaa > 10 !then
                y = 1;
            !end
        !else
            !for ?w = a, b, c !do
                u?w?x = 2;
            !end
        !end
    !end

!equations

    z = 100;

!variables{:main}

    a

"""

p, info = ir.parsers.preparser.from_string(source, context={"aaa": 1}, )
q = ir.parsers.algebraic.from_string(p, )

a, *_ = ir.sources.AlgebraicSource.from_string(source, context={"aaa": 1}, )





"""
Time frames for dynamic simulations
"""


#[

from __future__ import annotations

import numpy as _np

from .dates import SPAN_ELLIPSIS

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import (Callable, Iterable, )
    from .dataslates.main import Dataslate
    from .plans.simulation_plans import SimulationPlan
    from .dates import Period

#]


class Frame:
    """
    """
    #[

    __slots__ = (
        # Periods
        "start",
        "end",
        "simulation_end",
        # Columns
        "first",
        "last",
        "simulation_last",
        "slice",
        "zero_unanticipated_slice",
        "simulation_slice",
        "num_simulation_columns",
    )

    def __init__(
        self,
        start: Period,
        end: Period,
        simulation_end: Period | None = None,
    ) -> None:
        """
        """
        self.start = start
        self.end = end
        self.simulation_end = simulation_end
        self.first = None
        self.last = None
        self.simulation_last = None
        self.slice = None
        self.simulation_slice = None
        self.num_simulation_columns = None

    def resolve_columns(
        self,
        first_column_period: Period,
    ) -> None:
        """
        Given the time period of the first column of data array, determine the
        first and last columns of the frame
        """
        self.first = self.start - first_column_period
        self.last = self.end - first_column_period
        self.simulation_last = self.simulation_end - first_column_period
        self.slice = slice(self.first, self.last+1, )
        self.zero_unanticipated_slice = slice(self.first+1, None, )
        self.simulation_slice = slice(self.first, self.simulation_last+1, )
        self.num_simulation_columns = self.simulation_last - self.first + 1


class SplitFrame(Frame, ):
    """
    """
    #[

    def __repr__(self, ) -> str:
        """
        """
        return f"<SplitFrame {self.start}{SPAN_ELLIPSIS}{self.end}{SPAN_ELLIPSIS}{self.simulation_end}>"

    def prune_frame_data(
        self,
        frame_ds: Dataslate,
        unanticipated_qids: Iterable[int],
    ) -> None:
        """
        Remove unanticipated shocks after the first period of the frame
        """
        if self.start == self.simulation_end:
            return
        data = frame_ds.get_data_variant()
        data[unanticipated_qids, self.zero_unanticipated_slice] = 0

    def write_frame_data_to_main_dataslate(
        self,
        main_ds: Dataslate,
        frame_ds: Dataslate,
        unanticipated_shock_qids: Iterable[int],
    ) -> None:
        """
        Copy frame columns from frame_ds to main_ds
        """
        main_data = main_ds.get_data_variant(0, )
        frame_data = frame_ds.get_data_variant(0, )
        all_qids = range(frame_ds.num_names)
        regular_qids = tuple(
            qid for qid in all_qids
            if qid not in unanticipated_shock_qids
        )
        main_data[regular_qids, self.slice] = frame_data[regular_qids, self.slice]
        main_data[unanticipated_shock_qids, self.first] = frame_data[unanticipated_shock_qids, self.first]

    #]


class SingleFrame(Frame, ):
    """
    Single frame over the entire simulation span with all unanticipated data retained
    """
    #[

    def __init__(self, start: Period, end: Period, ) -> None:
        """
        """
        super().__init__(start, end, end, )

    def __repr__(self, ) -> str:
        """
        """
        return f"<SingleFrame {self.start}…{self.end}>"

    def prune_frame_data(
        self,
        frame_ds: Dataslate,
        model: Slatable,
    ) -> None:
        r"""
        No pruning, keep unanticipated shocks
        """
        pass

    def write_frame_data_to_main_dataslate(
        self,
        main_ds: Dataslate,
        frame_ds: Dataslate,
        unanticipated_shock_qids: Iterable[int],
    ) -> None:
        """
        """
        main_data = main_ds.get_data_variant(0, )
        frame_data = frame_ds.get_data_variant(0, )
        main_data[:, self.slice] = frame_data[:, self.slice]

    #]


def split_into_frames(
    model,
    dataslate: Dataslate,
    plan: SimulationPlan | None,
    get_simulation_end: Callable,
) -> tuple[bool, ...]:
    """
    """
    #[
    base_break_points = _populate_base_break_points(model, dataslate, plan, )
    return split_into_frames_by_breakpoints(
        base_break_points, dataslate, get_simulation_end,
    )
    #]


def setup_single_frame(
    model,
    dataslate: Dataslate,
    plan: SimulationPlan | None,
    **kwargs,
) -> tuple[SingleFrame]:
    """
    Return a tuple with a SingleFrame that spans the entire simulation period
    """
    #[
    base_periods = dataslate.base_periods
    first_dataslate_period = dataslate.periods[0]
    frame_start = base_periods[0]
    frame_end = base_periods[-1]
    simulation_end = frame_end
    frame = SingleFrame(frame_start, frame_end, )
    frame.resolve_columns(first_dataslate_period, )
    return (frame, )
    #]


def split_into_frames_by_breakpoints(
    base_break_points: tuple[bool, ...],
    dataslate: Dataslate,
    get_simulation_end: Callable,
) -> tuple[bool, ...]:
    """
    """
    #[
    base_periods = dataslate.base_periods
    break_periods = _get_break_periods_from_break_points(base_break_points, base_periods, )
    break_periods_next = break_periods[1:] + (base_periods[-1] + 1, )
    first_dataslate_period = dataslate.periods[0]
    frames = []
    for start, next_ in zip(break_periods, break_periods_next, ):
        end = next_ - 1
        new_frame = SplitFrame(start, end, get_simulation_end(start, end, ), )
        new_frame.resolve_columns(first_dataslate_period, )
        frames.append(new_frame, )
    return tuple(frames, )
    #]


def create_empty_base_break_points(dataslate: Dataslate, /, ) -> _np.ndarray:
    """
    """
    #[
    base_columns = dataslate.base_columns
    return _np.zeros((len(base_columns), ), dtype=bool, )
    #]


def _populate_base_break_points(
    model,
    dataslate: Dataslate | None,
    plan: SimulationPlan | None,
) -> tuple[Frame]:
    """
    """
    #[
    if not model.is_singleton:
        raise ValueError("Model must be a singleton")
    if not dataslate.is_singleton:
        raise ValueError("Dataslate must be a singleton")
    #
    base_break_points = create_empty_base_break_points(dataslate, )
    base_break_points[0] = True
    plannable = model.get_simulation_plannable()
    #
    # First, get breakpoints at the dates of unanticipated shocks
    name_to_row = dataslate.create_name_to_row()
    rows = tuple(
        name_to_row[name]
        for name in plannable.can_be_endogenized_unanticipated
    )
    base_columns = dataslate.base_columns
    if rows and base_columns:
        cutout = dataslate.get_data_variant(0, )[rows, :][:, base_columns]
        base_break_points = _update_break_points(base_break_points, cutout, )
    #
    # Then, get breakpoints at the dates of endogenized unanticipated shocks
    if plan is not None:
        base_periods = dataslate.base_periods
        cutout = plan.get_register_as_bool_array(
            "endogenized_unanticipated",
            plannable.can_be_endogenized_unanticipated,
            base_periods,
        )
        base_break_points = _update_break_points(base_break_points, cutout, )
    #
    return base_break_points
    #]


def _get_break_periods_from_break_points(
    base_break_points: Iterable[bool],
    base_periods: Iterable[Period],
    /,
) -> tuple[Period]:
    """
    """
    #[
    after_period = base_periods[-1] + 1
    zipped = zip(base_periods, base_break_points)
    return tuple(t for t, flag in zipped if flag)
    #]


def _update_break_points(base_break_points, new_array, ):
    """
    """
    #[
    if new_array is None:
        return base_break_points
    new_break_points = _np.any(_np.isfinite(new_array) & (new_array != 0), axis=0, )
    return base_break_points | new_break_points
    #]





import irispie as ir

class X:
    pass

x = X()
x.simulate_can_be_exogenized = ["a", "b", "c"]
x.simulate_can_be_endogenized = ["res_a", "res_b", "res_c"]

rng = ir.qq(2020,1) >> ir.qq(2021,4)
p = ir.PlanSimulate(x, rng, )
p.exogenize(ir.qq(2020,2), "a", when_data=True, transform="log", )
p.endogenize(ir.qq(2020,2), "res_a", )
print(p.pretty)





import irispie as ir
import random as rn

# Generate random values
values = tuple(rn.gauss() for _ in range(16))

db = ir.Databox()

# Create a quarterly time series
db["q"] = ir.Series(start=ir.qq(2020,1), values=values, )

# Create a undated time series
db["i"] = ir.Series(start=ir.ii(1), values=values, )

# Plot the quarterly time series with observations and labels in the middle of
# the period
db["q"].plot(
    date_axis_mode="period",
    highlight=(ir.qq(2020,1), ir.qq(2021,2)),
    xline=ir.qq(2020,3),
    show_figure=False,
)

# Plot the quarterly time series with observations and labels at the start of
# the period
db["q"].plot(
    date_axis_mode="instant",
    highlight=(ir.qq(2020,1), ir.qq(2021,2)),
    xline=ir.qq(2020,3),
    show_figure=False,
)

# Plot the undated time series with observations and labels in the middle of
# the period
db["i"].plot(
    date_axis_mode="period",
    highlight=(ir.ii(1), ir.ii(6)),
    xline=ir.ii(3),
    show_figure=False,
)

# Plot the undated time series with observations and labels at the start of
# the period
db["i"].plot(
    date_axis_mode="instant",
    highlight=(ir.ii(1), ir.ii(6)),
    xline=ir.ii(3),
    show_figure=False,
)





import irispie as ir
import re
import itertools as it


# source = """
# 
# !equations
# 
#     !for ?1 = <<level_one>> !do
#         !for ?x = <level_two.keys()> !do
#             !for ?(z) = <level_two.values()> !do
#                 ?1?x?(z) = 1;
#             !end
#         !end
#     !end
# 
# """
# 
# context = {
#     "level_one": ["a", "b", "c", ],
#     "level_two": {1: "A", 2: "B", 3: "C", },
# }
# 
# _, info = ir.parsers.preparser.from_string(source, context=context, )
# 
# actual_list = re.findall(r'\w+(?= *=)', info['preparsed_source'])
# expected_list = [
#     f"{x}{y}{z}"
#     for x, y, z in it.product(
#         context["level_one"],
#         context["level_two"].keys(),
#         context["level_two"].values(),
#     )
# ]
# 
# assert expected_list == actual_list


#-------------------------------------------------------------------------------



source = """

!equations

#     !for ?(stock) = <stocks> !do
# 
#         !for ?(flow) = <pos_flows> !do
#             ?(stock)_?(flow) = 0;
#         !end
# 
#         !for ?(flow) = <neg_flows> !do
#             ?(stock)_?(flow) = 0;
#         !end
# 
#         stock_?(stock) = ...
#             !for ?(flow) = <pos_flows> !do + ?(stock)_?(flow) !end ...
#             !for ?(flow) = <neg_flows> !do - ?(stock)_?(flow) !end ...
#         ;
# 
#     !end

    !for ?(a) = < stock_aggs.keys() > !do
        #!
        #! Stock-flow equation ?{a}
        #!
        "Stock-flow equation ?{a}"
        stock_?(a) = ...
            !for ?(c) = < stock_aggs["?(a)"] > !do
                + stock_?(c) + ...
            !end
        ;
    !end

"""

context = {
    "stock_aggs": {
        "mlt": ["mt", "lt", ],
        "tot": ["mlt", "st", ],
    },
    "stocks": [ "st", "mt", "lt",],
    "pos_flows": [ "new", "rev" ],
    "neg_flows": [ "rep", ],
}

_, info = ir.parsers.preparser.from_string(source, context=context, )

print(info['preparsed_source'])





import irispie as ir


db = ir.Databox()
db['aa'] = ir.Series(num_variants=2, dates=ir.qq(2020,1,...,2020,4), values=[(1,2,3,4), 0], )
db['bb'] = 3
db['cc'] = [100, 200]

ds = ir.Dataslate.from_databox(
    db, tuple(db.keys()), ir.qq(2019,1,...,2020,4),
    num_variants=3,
    scalar_names=('bb', 'cc'),
)

db2 = ir.Databox()
db2['dd'] = 'abc'
db2 = ds.to_databox(target_databox=db2, )







import irispie as ir
import numpy as _np
import scipy as _sp
import sys
from irispie.incidences import permutations as _permutations

source_code = """

!equations

    "aaa"
    a = 0.8*a[-1];
    y = b{-1};
    roc(b) = 0.5*a + c[-2] + d[+2];
    s === a + b + c + y;
    u = u[-1];

"""

z = ir.Sequential.from_string(source_code, )

sys.exit()
qid_to_name = z.create_qid_to_name()

d = ir.Databank()
im = z.incidence_matrix
lhs_names = z.lhs_names

# #ims = _sp.sparse.csc_matrix(im)
# ims = im
# lu = _sp.sparse.linalg.splu(ims, permc_spec="MMD_ATA")
# im1 = im;
# im1 = im1[lu.perm_r, :]
# im1 = im1[:, lu.perm_c]
# print(im1)

s = z
s.is_sequential
rows = s.sequentialize()


# (rows, columns), _, info = _permutations.sequentialize(im)
# im = im[rows, :]
# im = im[:, rows]
# print(im)
# 

start_sim = ir.yy(2020)
end_sim = ir.yy(2024)
sim_rng = ir.yy(2020) >> ir.yy(2024)
db_rng = start_sim-10 >> end_sim+2
db = ir.Databank.for_simulation(s, db_rng, func=_np.random.uniform)
for n in s.res_names:
    db[n][...] = 0

sim = s.simulate(db, sim_rng, )





r"""
"""


#[

from __future__ import annotations

from typing import NamedTuple

#]


class Dimensions(NamedTuple, ):
    r"""
    """
    #[

    num_endogenous: int | None = None
    order: int | None = None
    has_intercept: bool = True
    num_exogenous: int = 0

    @property
    def num_nonendogenous(self, ) -> int:
        return self.num_exogenous + int(self.has_intercept)

    @property
    def num_lagged_endogenous(self, ) -> int:
        return self.num_endogenous * self.order

    @property
    def num_rhs(self, ) -> int:
        return self.num_lagged_endogenous + self.num_nonendogenous

    @property
    def has_exogenous(self, ) -> bool:
        return self.num_exogenous > 0

    #]






import irispie as _ir
import json as _js


source = """
    !variables
        !for ? = <range(1000)> !do
            x?
        !end
    !equations
        !for ?i = <range(1000)> !do
            x?i = !for ?j = <range(1,1000,70)> !do + x?j[-3] !end;
        !end
"""

m = _ir.Model.from_string(source, linear=True, )



d = {
    n: list(range(100))
    for n in ("x"+str(i) for i in range(1000))
}






import irispie as ir
import pytest
import os


_THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def test_progress_bar():
    log_file = os.path.join(_THIS_DIR, "progress_bar_test.log")
    with open(log_file, "wt", ) as fid:
        p = ir.ProgressBar(num_steps=10, output_stream=fid, )
        p.start()
        for i in range(10):
            p.increment()
        p.finish()
    with open(log_file, "rt") as fid:
        log = fid.read()
    assert log[-1] == "\n"
    assert log[-2] == "s"





import pytest
import irispie as ir
import random as rn


rn.seed(0)
_ROUND_TO = 5


@pytest.mark.parametrize(
    ["func"], [ ("pct", ), ("log", ), ("diff", ), ("roc", ), ("diff_log", ) ],
)
def test_simulate(func: str ):

    source = r"""
    !equations
        {{func}}(x) = {{func}}(y);
    """

    m = ir.Sequential.from_string(
        source,
        context={"func": func},
    )

    start_sim = ir.qq(2020,1)
    end_sim = ir.qq(2025,4)

    db = ir.Databox()
    db["y"] = ir.Series(
        periods=start_sim-1>>end_sim,
        func=lambda: 2 + rn.uniform(0, 1),
    )
    db["x"] = db["y"](start_sim-1)

    s = m.simulate(db, start_sim>>end_sim, )

    assert ir.round(s["x"], _ROUND_TO) == ir.round(s["y"], _ROUND_TO)




"""
Iteration printers for dynamic period-by-period systems
"""


#[

from __future__ import annotations

import neqs as _nq

from .. import quantities as _quantities

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable
    from ..equations import Equation

#]


_ITER_PRINTER_COLUMNS = (
    "counter",
    "func_norm",
    "step_length",
    "jacob_status",
    "worst_diff_x",
    "worst_func",
)


def create_iter_printer(
    equations: Iterable[Equation],
    qids: Iterable[str],
    qid_to_logly: dict[str, bool],
    qid_to_name: dict[str, str],
    **kwargs,
) -> None:
    """
    """
    equation_strings = _get_equation_strings(equations, )
    quantity_strings = _get_quantity_strings(qids, qid_to_logly, qid_to_name, )
    return _nq.IterPrinter(
        equation_strings=equation_strings,
        quantity_strings=quantity_strings,
        columns=_ITER_PRINTER_COLUMNS,
        **kwargs,
    )


def _get_equation_strings(equations: Iterable[Equation], /, ) -> tuple[str, ...]:
    return tuple(i.human for i in equations)


def _get_quantity_strings(
    qids: Iterable[str],
    qid_to_logly: dict[str, bool],
    qid_to_name: dict[str, str],
    /,
) -> tuple[str, ...]:
    return tuple(
        _quantities.wrap_logly(qid_to_name[i], qid_to_logly.get(i, False), )
        for i in qids
    )




"""
Implement SimulationPlannableProtocol and SteadyPlannableProtocol
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, )

from .. import quantities as _quantities

if TYPE_CHECKING:
    from .main import (Simultaneous, )
#]


class _SimulationPlannable:
    """
    """
    #[

    def __init__(
        self,
        model,
        **kwargs,
    ) -> None:
        """
        """
        curr_xi_qids, *_ = model.solution_vectors.get_curr_transition_indexes()
        qid_to_name = model.create_qid_to_name()
        can_be_exogenized = tuple( qid_to_name[qid] for qid in curr_xi_qids )
        #
        self.can_be_exogenized_anticipated = can_be_exogenized
        self.can_be_exogenized_unanticipated = can_be_exogenized
        #
        def _get_names(kind: _quantities.QuantityKind, ):
            return tuple(_quantities.generate_quantity_names_by_kind(
                model.quantities, kind=kind,
            ))
        #
        self.can_be_endogenized_anticipated \
            = _get_names(_quantities.ANTICIPATED_SHOCK, )
        #
        self.can_be_endogenized_unanticipated \
            = _get_names(_quantities.UNANTICIPATED_SHOCK, )

    #]


class _SteadyPlannable:
    """
    """
    #[

    def __init__(
        self,
        model: Simultaneous,
        is_flat: bool,
        **kwargs,
    ) -> None:
        """
        """
        def _get_names(kind: _quantities.QuantityKind, ):
            return tuple(_quantities.generate_quantity_names_by_kind(
                model.quantities, kind=kind,
            ))
        #
        self.can_be_exogenized = _get_names(_quantities.ENDOGENOUS_VARIABLE, )
        self.can_be_endogenized = _get_names(_quantities.PARAMETER, )
        self.can_be_fixed_level = self.can_be_exogenized
        if is_flat:
            self.can_be_fixed_change = ()
        else:
            self.can_be_fixed_change = self.can_be_exogenized

    #]


class Inlay:
    """
    """
    #[

    def get_simulation_plannable(self, **kwargs, ) -> _SimulationPlannable:
        """
        """
        return _SimulationPlannable(self, **kwargs, )

    def get_steady_plannable(self, **kwargs, ) -> _SteadyPlannable:
        model_flags = self.resolve_flags(**kwargs, )
        return _SteadyPlannable(self, model_flags.is_flat, )

    #]




"""
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp

from . import systems as _systems
from . import solutions as _solutions

#]


def solve_steady_linear_flat(
    sys: _systems.System,
    /,
) -> tuple[_np.ndarray, _np.ndarray, _np.ndarray, _np.ndarray]:
    """
    """
    #[
    left_div = _solutions.left_div
    vstack = _np.vstack
    hstack = _np.hstack
    concatenate = _np.concatenate
    #
    A, B, C, F, G, H = sys.A, sys.B, sys.C, sys.F, sys.G, sys.H
    #
    # A @ Xi + B @ Xi{-1} + C = 0
    # F @ Y + G @ Xi + H = 0
    #
    # Xi = -pinv(A + B) @ C
    Xi = left_div(-(A + B), C, )
    dXi = _np.zeros(Xi.shape)
    #
    # Y = -pinv(F) @ (G @ Xi + H)
    Y = left_div(-F, G @ Xi + H, )
    dY = _np.zeros(Y.shape)
    #
    return Xi, Y, dXi, dY
    #]


def solve_steady_linear_nonflat(
    sys: _systems.System,
    /,
) -> tuple[_np.ndarray, _np.ndarray, _np.ndarray, _np.ndarray]:
    """
    """
    #[
    """
    """
    left_div = _solutions.left_div
    vstack = _np.vstack
    hstack = _np.hstack
    concatenate = _np.concatenate
    #
    A, B, C, F, G, H = sys.A, sys.B, sys.C, sys.F, sys.G, sys.H
    num_y = F.shape[0]
    k = 1
    #
    # A @ Xi + B @ Xi{-1} + C = 0:
    # -->
    # A @ Xi + B @ (Xi - dXi) + C = 0
    # A @ (Xi + k*dXi) + B @ (Xi + (k-1)*dXi) + C = 0
    #
    AB = vstack((
        hstack(( A + B, 0*A + (0-1)*B )),
        hstack(( A + B, k*A + (k-1)*B )),
    ))
    CC = concatenate((C, C, ))
    # Xi_dXi = -pinv(AB) @ CC
    Xi_dXi = left_div(-AB, CC, )
    #
    # F @ Y + G @ Xi + H = 0:
    # -->
    # F @ Y + G @ Xi + H = 0
    # F @ (Y + k*dY) + G @ (Xi + k*dXi) + H = 0
    #
    FF = vstack((
        hstack(( F, 0*F )),
        hstack(( F, k*F )),
    ))
    GG = vstack((
        hstack(( G, 0*G )),
        hstack(( G, k*G )),
    ))
    HH = concatenate((H, H, ))
    # Y_dY = -pinv(FF) @ (GG @ Xi_dXi + HH)
    Y_dY = left_div(-FF, GG @ Xi_dXi + HH, )
    #
    # Separate levels and changes
    #
    num_xi = A.shape[1]
    num_y = F.shape[1]
    Xi, dXi = (
        Xi_dXi[0:num_xi, ...],
        Xi_dXi[num_xi:, ...],
    )
    Y, dY = (
        Y_dY[0:num_y, ...],
        Y_dY[num_y:, ...]
    )
    #
    return Xi, Y, dXi, dY
    #]






class Model:
    def __init__(self, context, ):
        self.context = context
        self.func_string = "def func(a, b, ): return a * cdf(a, b)"
        self._compile_function()

    def _compile_function(self, ):
        context = {"__builtins__": None, } | self.context
        exec(self.func_string, context, )
        self.func = context["func"]

    def eval(self, a, b, ):
        return self.func(a, b, )

    def __getstate__(self, ):
        return {"context": self.context, "func_string": self.func_string, }

    def __setstate__(self, state):
        self.context = state["context"]
        self.func_string = state["func_string"]
        self._compile_function()




import irispie as ir
from irispie.databoxes import _imports2
import sys

d = ir.Databox()
d["a"] = ir.Series(dates=ir.qq(2020,1)>>ir.qq(2024,4), values=range(20))
d["b"] = d["a"] + 10
d["c"] = (d["a"] + 100)[-10]

d["x"] = ir.Series()

d.apply(
    lambda x: ir.aggregate(x, ir.Freq.YEARLY),
    source_names=["a", "b", "c"],
    target_names=lambda n: n + "_yearly",
)

d.apply(
    lambda x: x.redate(ir.ii(0)),
    source_names=["a", "b", "c"],
    target_names=lambda n: n + "_int",
    in_place=True,
)
sys.exit()

info = d.to_sheet("test_export_to_sheet.csv", )
e = ir.Databox.from_sheet("test_export_to_sheet.csv", )


for k, v in d.items():
    assert all(v.get_data(...) == e[k].get_data(...))


d.to_sheet(
    "test_export_to_sheet_monthly.csv",
    #frequency_range={ir.Freq.MONTHLY: ir.mm(2020,1)>>ir.mm(2022,12)},
    frequency_range={ir.Freq.MONTHLY: ...},
)

e1 = ir.Databox.from_sheet("test_export_to_sheet_monthly.csv", )





"""
First-order terminators
========================

"""


#[
from __future__ import annotations

from types import (SimpleNamespace, )
import itertools as _it
import numpy as _np
import scipy as _sp
import warnings as _wa

from ..incidences import main as _incidences
from ..incidences.main import Token
from .. import equations as _equations
from ..equations import Equation
from . import simulators as _simulators
from ..aldi.maps import ArrayMap

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Callable, )
    from collections.abc import (Iterable, Sequence, )
    from .simulators import FordSimulatableProtocol
#]


class Terminator:
    """
    First-order terminator class
    """
    #[

    def __init__(
        self,
        simulatable_v: FordSimulatableProtocol,
        columns_simulated: Sequence[int],
        wrt_equations: Iterable[Equation],
    ) -> None:
        """
        """
        last_simulation = columns_simulated[-1]
        first_terminal = last_simulation + 1
        max_lead = simulatable_v.max_lead
        terminal_columns = tuple(range(first_terminal, first_terminal + max_lead, ), )
        num_terminal_columns = len(terminal_columns)
        #
        solution = simulatable_v._get_singleton_solution(deviation=False, )
        vec = simulatable_v._get_dynamic_solution_vectors()
        qid_to_logly = simulatable_v.create_qid_to_logly()
        #
        curr_xi_qids, curr_xi_indexes = vec.get_curr_transition_indexes()
        num_curr_xi_qids = len(curr_xi_qids)
        #
        tokens_from_equations = (
            i for i in _equations.generate_all_tokens_from_equations(wrt_equations, )
            if i.qid in curr_xi_qids
        )
        curr_xi_qid_to_max_shift = _incidences.get_some_shift_by_quantities(tokens_from_equations, max, )
        to_be_zipped = (
            (inx, Token(qid, col, ), )
            for inx, (col, qid, ) in enumerate(_it.product(terminal_columns, curr_xi_qids, ))
            if col <= last_simulation + curr_xi_qid_to_max_shift[qid]
        )
        terminal_column_index, terminal_wrt_spots = zip(*to_be_zipped)
        #
        #
        T = solution.T
        K = solution.K
        #
        # curr_TT := [ T; T @ T; ... ] with current dated rows only
        # curr_KK := [ K; T @ K + K; ... ] with current dated rows only
        curr_TT = []
        curr_KK = []
        cum_T = _np.eye(*T.shape, dtype=T.dtype, )
        cum_K = _np.zeros(K.shape, dtype=K.dtype, )
        for i in range(max_lead, ):
            cum_T = T @ cum_T
            cum_K = T @ cum_K + K
            curr_TT.append(cum_T[curr_xi_indexes, ...])
            curr_KK.append(cum_K[curr_xi_indexes, ...])
        #
        self.terminal_wrt_spots = terminal_wrt_spots
        self._transition_vector = vec.transition_variables
        self._logly_rows = tuple( qid for qid, status in qid_to_logly.items() if status )
        self._first_terminal = first_terminal
        self._max_lead = max_lead
        self._curr_xi_qids = curr_xi_qids
        self._curr_TT = _np.vstack(curr_TT, )
        self._curr_KK = _np.hstack(curr_KK, )
        self._terminal_columns = terminal_columns
        self._terminal_column_index = terminal_column_index
        self._terminit_spots = tuple( Token(i.qid, last_simulation + i.shift, ) for i in vec.transition_variables )
        self._num_terminal_wrt_spots = len(terminal_wrt_spots)
        self.terminal_jacobian_map = None
        self._terminal_jacobian_map_completed = False

    def create_terminal_jacobian_map(self, wrt_spots: Iterable[Token], ) -> None:
        """
        """
        lhs_columns = []
        rhs_columns = []
        wrt_spot_to_index = { spot: inx for inx, spot in enumerate(wrt_spots, ) }
        for rhs_column, spot in enumerate(self._terminit_spots, ):
            try:
                index = wrt_spot_to_index[spot]
            except KeyError:
                continue
            lhs_columns.append(index)
            rhs_columns.append(rhs_column)
        self.terminal_jacobian_map = ArrayMap(
            lhs=([], lhs_columns),
            rhs=([], rhs_columns),
        )

    def terminate_simulation(self, data_array: _np.ndarray, /, ) -> None:
        """
        """
        logly_rows = self._logly_rows
        first_terminal = self._first_terminal
        transition_vector = self._transition_vector
        max_lead = self._max_lead
        terminal_columns = self._terminal_columns
        curr_xi_qids = self._curr_xi_qids
        curr_TT = self._curr_TT
        curr_KK = self._curr_KK
        #
        data_array[logly_rows, :] = _np.log(data_array[logly_rows, :])
        terminit_xi = _simulators.get_init_xi(data_array, transition_vector, first_terminal, )
        terminal_curr_xi_flat = curr_TT @ terminit_xi + curr_KK
        terminal_curr_xi = terminal_curr_xi_flat.reshape(-1, max_lead, order="F", )
        data_array[_np.ix_(curr_xi_qids, terminal_columns)] = terminal_curr_xi
        data_array[logly_rows, :] = _np.exp(data_array[logly_rows, :])

    def terminate_jacobian(self, jacobian_outcome: _np.ndarray, /, ) -> _np.ndarray:
        """
        """
        num_terminal_wrt_spots = len(self.terminal_wrt_spots)
        terminal_column_index = self._terminal_column_index
        #
        terminal_jacobian_outcome = jacobian_outcome[:, -num_terminal_wrt_spots:]
        regular_jacobian_outcome = jacobian_outcome[:,:-num_terminal_wrt_spots].copy()

        if not self._terminal_jacobian_map_completed:
            coo = terminal_jacobian_outcome.tocoo()
            nnz_rows = sorted(set(coo.row))
            _complete_terminal_jacobian_map(self.terminal_jacobian_map, nnz_rows, )
            self._terminal_jacobian_map_completed = True

        curr_TT = self._curr_TT[terminal_column_index, :]
        add_to_regular_jacobian_outcome = terminal_jacobian_outcome @ curr_TT

        _wa.filterwarnings("ignore", category=_sp.sparse.SparseEfficiencyWarning, )
        regular_jacobian_outcome[self.terminal_jacobian_map.lhs] += add_to_regular_jacobian_outcome[self.terminal_jacobian_map.rhs]
        _wa.filterwarnings("default", category=_sp.sparse.SparseEfficiencyWarning, )

        return regular_jacobian_outcome

    #]


def _complete_terminal_jacobian_map(
    self: ArrayMap,
    nnz_rows: Iterable[int],
    /,
) -> None:
    """
    """
    lhs_columns = self.lhs[1]
    rhs_columns = self.rhs[1]
    lhs_grid = _np.ix_(nnz_rows, lhs_columns, )
    rhs_grid = _np.ix_(nnz_rows, rhs_columns, )
    self.lhs = lhs_grid
    self.rhs = rhs_grid




"""
"""


#[

from __future__ import annotations

import itertools as _it

from ..conveniences import descriptions as _descriptions
from .. import quantities as _quantities
from ..quantities import Quantity, QuantityKind
from ..dataslates import Dataslate
from ..fords.descriptors import SolutionVectors
from ..incidences.main import Token

from ._dimensions import Dimensions

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable

#]


class Invariant(
    _descriptions.DescriptionMixin,
):
    r"""
    """
    #[

    __slots__ = (
        "quantities",
        "dimensions",
        "solution_vectors",
    )

    def __init__(
        self,
        endogenous_names: Iterable[str],
        exogenous_names: Iterable[str] | None = None,
        order: int = 1,
        intercept: bool = True,
    ) -> None:
        """
        """
        endogenous_names = tuple(endogenous_names)
        exogenous_names = tuple(exogenous_names) if exogenous_names else ()
        self.quantities = _create_quantities(endogenous_names, exogenous_names, )
        self.dimensions = Dimensions(
            num_endogenous=len(endogenous_names),
            order=order,
            has_intercept=intercept,
            num_exogenous=len(exogenous_names),
        )
        self._populate_solution_vectors()

    def _get_some_qids(self, kind: QuantityKind, ) -> tuple[int, ...]:
        r"""
        """
        return tuple(_quantities.generate_qids_by_kind(self.quantities, kind, ))

    def get_endogenous_qids(self, ) -> tuple[int, ...]:
        r"""
        """
        return self._get_some_qids(QuantityKind.TRANSITION_VARIABLE, )

    def get_residual_qids(self, ) -> tuple[int, ...]:
        r"""
        """
        return self._get_some_qids(QuantityKind.UNANTICIPATED_SHOCK, )

    def get_exogenous_qids(self, ) -> tuple[int, ...]:
        r"""
        """
        return self._get_some_qids(QuantityKind.EXOGENOUS_VARIABLE, )

    def _populate_solution_vectors(self, ) -> None:
        r"""
        """
        order = self.dimensions.order
        num_lagged_endogenous = self.dimensions.num_lagged_endogenous
        endogenous_qids = self.get_endogenous_qids()
        residual_qids = self.get_residual_qids()
        measurement_tokens = tuple(
            Token(qid=qid, shift=0, )
            for qid in endogenous_qids
        )
        transition_tokens = tuple(
            Token(qid=qid, shift=shift, )
            for shift, qid in _it.product(range(order), endogenous_qids, )
        )
        residual_tokens = tuple(
            Token(qid=qid, shift=0, )
            for qid in residual_qids
        )
        self.solution_vectors = SolutionVectors(
            transition_variables=transition_tokens,
            unanticipated_shocks=residual_tokens,
            anticipated_shocks=(),
            measurement_variables=measurement_tokens,
            measurement_shocks=(),
            true_initials=(True, ) * num_lagged_endogenous,
        )

    #]


def _create_quantities(
    endogenous_names: Iterable[str],
    exogenous_names: Iterable[str] | None = None,
) -> tuple[Quantity, ...]:
    """
    """
    quantities = []
    def append_quantities(names, kind, ):
        for n in names:
            quantities.append(Quantity(id=len(quantities), human=n, kind=kind, ))
    #
    endogenous_names = tuple(endogenous_names)
    exogenous_names = tuple(exogenous_names or ())
    residual_names = tuple(
        _residual_name_from_endogenous_name(name)
        for name in endogenous_names
    )
    #
    append_quantities(endogenous_names , QuantityKind.TRANSITION_VARIABLE, )
    append_quantities(exogenous_names, QuantityKind.EXOGENOUS_VARIABLE, )
    append_quantities(residual_names, QuantityKind.UNANTICIPATED_SHOCK, )
    #
    return tuple(quantities, )


_RESIDUAL_NAME_PREFIX = "res_"


def _residual_name_from_endogenous_name(n: str) -> str:
    r"""
    """
    return f"{_RESIDUAL_NAME_PREFIX}{n}"




"""
Dynamic nonlinear stacked-time simulator
"""


#[

from __future__ import annotations

import warnings as _wa
import numpy as _np
import scipy as _sp
import itertools as _it
import neqs as _nq

from ..simultaneous import main as _simultaneous
from ..plans.simulation_plans import SimulationPlan
from ..dataslates.main import Dataslate
from ..incidences import main as _incidences
from ..incidences.main import Token
from .. import quantities as _quantities
from .. import equations as _equations
from .. import frames as _frames
from .. import dates as _dates
from ..dates import Span
from ..frames import SingleFrame
from .. import wrongdoings as _wrongdoings
from ..fords import simulators as _ford_simulators
from ..fords.terminators import Terminator

from . import _evaluators as _evaluators
from . import _iter_printers as _iter_printers

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from numbers import Real
    from typing import Callable, Any, Literal
    from ..dates import Period
    from ..simultaneous.main import Simultaneous
    from ..frames import Frame

#]


# Simulator protocol requirements
METHOD_NAME = "stacked_time"
def create_frames(): ...
def simulate_initial_guess(): ...
def simulate_frame(): ...


_DEFAULT_FALLBACK_VALUE = 1/9


_RELEVANT_REGISTER_NAMES = (
    "exogenized_anticipated",
    "endogenized_anticipated",
    "exogenized_unanticipated",
    "endogenized_unanticipated",
)


def create_frames(
    model_v: Simultaneous,
    dataslate_v: Dataslate,
    plan: SimulationPlan | None,
    **kwargs,
) -> tuple[Frame, ...]:
    """
    """
    #[
    base_end = dataslate_v.base_periods[-1]
    return _frames.split_into_frames(
        model_v, dataslate_v, plan,
        get_simulation_end=lambda *_, : base_end,
    )
    #]


def simulate_initial_guess(
    model_v: Simultaneous,
    dataslate_v: Dataslate,
    plan: SimulationPlan | None,
    *,
    initial_guess: Literal["first_order", "data"] = "first_order",
    **kwargs,
) -> None:
    """
    """
    _INITIAL_GUESS_SIMULATOR[initial_guess](model_v, dataslate_v, )


def simulate_frame(
    model_v: Simultaneous,
    frame_ds: Dataslate,
    *,
    frame: Frame,
    input_data_array: _np.ndarray,
    plan: SimulationPlan | None,
    simulation_header: str,
    return_info: bool = False,
    # Method specific settings
    solver_settings: dict[str, Any] | None = None,
    when_fails: Literal["critical", "error", "warning", "silent"] = "error",
    when_missing: Literal["critical", "error", "warning", "silent"] = "error",
    fallback_value: Real = _DEFAULT_FALLBACK_VALUE,
    terminal: Literal["data", "first_order", ] = "first_order",
    _precatch_missing: Callable | None = None,
    **kwargs,
) -> dict[str, Any]:
    """
    """
    #[

    solver_settings = solver_settings or {}

    when_missing_stream = \
        _wrongdoings.STREAM_FACTORY[when_missing] \
        (f"These values are missing at the start of the {frame} simulation: ")

    all_quantities = model_v.get_quantities()
    qid_to_logly = _quantities.create_qid_to_logly(all_quantities, )
    qid_to_name = model_v.create_qid_to_name()
    name_to_qid = model_v.create_name_to_qid()

    endogenous_quantities = model_v.get_quantities(kind=_quantities.ENDOGENOUS_VARIABLE, )
    endogenous_qids = tuple(i.id for i in endogenous_quantities)
    wrt_equations = model_v.get_dynamic_equations(kind=_equations.ENDOGENOUS_EQUATION, )

    if len(wrt_equations) != len(endogenous_qids):
        raise _wrongdoings.IrisPieCritical(
            f"Number of endogenous equations {len(wrt_equations)} "
            f"does not match number of endogenous quantities {len(endogenous_qids)}"
        )

    if not wrt_equations:
        success = True
        return success

    periods = frame_ds.periods
    columns_to_run = tuple(range(frame.first, frame.simulation_last+1, ))
    periods_to_run = tuple(periods[i] for i in columns_to_run)

    wrt_spots, exogenized_spots = _get_wrt_spots(
        plan=plan,
        endogenous_qids=endogenous_qids,
        columns_to_run=columns_to_run,
        periods_to_run=periods_to_run,
        name_to_qid=name_to_qid,
    )

    terminator = None
    needs_terminator = terminal == "first_order" and model_v.max_lead
    if needs_terminator:
        terminator = Terminator(model_v, columns_to_run, wrt_equations, )
        terminator.create_terminal_jacobian_map(wrt_spots, )

    evaluator = _evaluators.create_evaluator(
        wrt_spots=wrt_spots,
        columns_to_eval=columns_to_run,
        wrt_equations=wrt_equations,
        all_quantities=all_quantities,
        terminator=terminator,
        context=model_v.get_context(),
    )

    data = frame_ds.get_data_variant()

    if _precatch_missing is not None:
        _precatch_missing(data, wrt_spots, )

    _catch_missing(
        data=data,
        wrt_spots=wrt_spots,
        frame=frame,
        qid_to_name=qid_to_name,
        fallback_value=fallback_value,
        when_missing_stream=when_missing_stream,
        periods=periods,
    )

    _copy_exogenized_data_to_frame_data(data, exogenized_spots, input_data_array, )

    iter_printer = _iter_printers.create_iter_printer(
        equations=wrt_equations,
        qids=tuple(tok.qid for tok in wrt_spots),
        qid_to_logly=qid_to_logly,
        qid_to_name=qid_to_name,
        custom_header=simulation_header,
        **solver_settings,
    )

    init_guess = evaluator.get_init_guess(data, )

    final_guess, exit_status = _nq.damped_newton(
        eval_func=evaluator.eval_func,
        eval_jacob=evaluator.eval_jacob,
        init_guess=init_guess,
        iter_printer=iter_printer,
        args=(data, ),
        **solver_settings,
    )

    evaluator.update(final_guess, data, )
    return exit_status

    #]


def _get_wrt_spots(
    plan: SimulationPlan | None,
    endogenous_qids: Iterable[int],
    columns_to_run: Iterable[int],
    periods_to_run: Iterable[Period],
    name_to_qid: dict[str, int],
) -> tuple[tuple[Token, ...], set[Token, ...]]:
    """
    """
    wrt_spots = tuple(
        Token(qid, column, )
        for column, qid in _it.product(columns_to_run, endogenous_qids, )
    )
    if plan is None:
        return wrt_spots, None,
    #
    registers_as_bool_arrays = plan.get_registers_as_bool_arrays(
        periods=periods_to_run,
        register_names=_RELEVANT_REGISTER_NAMES,
    )
    row_names_in_registers = {
        n: tuple(plan.get_register_by_name(n, ).keys())
        for n in _RELEVANT_REGISTER_NAMES
    }
    base_first = columns_to_run[0]
    def spots_from_register(register_name: str, columns_to_run: Iterable[int], ) -> set[Token]:
        return set(
            Token(name_to_qid[n], shift, )
            for row, n in enumerate(row_names_in_registers[register_name])
            for column, shift in enumerate(columns_to_run)
            if registers_as_bool_arrays[register_name][row, column]
        )
    exogenized_spots = (
        spots_from_register("exogenized_anticipated", columns_to_run, )
        | spots_from_register("exogenized_unanticipated", columns_to_run[0:1], )
    )
    endogenized_spots = (
        spots_from_register("endogenized_anticipated", columns_to_run, )
        | spots_from_register("endogenized_unanticipated", columns_to_run[0:1], )
    )
    wrt_spots = tuple(sorted(
        set(wrt_spots)
        .difference(exogenized_spots)
        .union(endogenized_spots)
    ))
    return wrt_spots, exogenized_spots


def _catch_missing(
    data: _np.ndarray,
    wrt_spots: Iterable[Token],
    frame: Frame,
    qid_to_name: dict[int, str],
    when_missing_stream: _wrongdoings.Stream,
    periods: Iterable[Period],
    fallback_value: Real,
) -> None:
    """
    """
    #[
    missing = _np.isnan(data)
    if not missing.any():
        return
    for qid, column in wrt_spots:
        if not missing[qid, column]:
            continue
        when_missing_stream.add(f"{qid_to_name[qid]}[{periods[column]}]", )
        data[qid, column] = fallback_value
    when_missing_stream._raise()
    #]


def _catch_fail(
    root_final: _sp.optimize.OptimizeResult,
    when_fails: Literal["error", "warning", "silent"],
) -> None:
    """
    """
    raise ValueError()


def _simulate_initial_guess_first_order(
    model_v: Simultaneous,
    frame_ds: DataSlate,
) -> None:
    """
    Simulate initial guess by running a first-order simulation without any
    shocks or conditioning
    """
    frame = SingleFrame(frame_ds.base_periods[0], frame_ds.base_periods[-1], )
    frame.resolve_columns(frame_ds.periods[0], )
    _ford_simulators.simulate_flat(
        model_v, frame_ds, frame,
        deviation=False,
        ignore_shocks=True,
    )


def _simulate_initial_guess_data(
    *args, **kwargs,
) -> None:
    """
    Initial guess is the data
    """
    pass


_INITIAL_GUESS_SIMULATOR = {
    "first_order": _simulate_initial_guess_first_order,
    "data": _simulate_initial_guess_data,
}


def _copy_exogenized_data_to_frame_data(
    data: _np.ndarray,
    exogenized_spots: set[Token],
    input_data_array: _np.ndarray,
) -> None:
    """
    """
    if not exogenized_spots:
        return
    indexes = tuple(zip(*exogenized_spots))
    data[indexes] = input_data_array[indexes]





import irispie as ir


db = ir.Databox()

db["x"] = ir.Series(start=ir.qq(2020,1), values=range(20), )
db["y"] = ir.Series(start=ir.qq(2015,1), values=range(60), )


# db["x"].plot()
span = ir.qq(2015,1,...,2030,4)

info = db["x"].plot(
    span=span,
    return_info=True,
    freeze_span=True,
    highlight=ir.qq(2020,1)>>ir.qq(2030,4),
    show_figure=False,
)


